"""TcEx Framework Module"""

# standard library
import logging
from collections.abc import Generator
from typing import Any, Self

# third-party
from pydantic.fields import ModelField  # TYPE-CHECKING

# first-party
from tcex.input.field_type.exception import InvalidEmptyValue, InvalidLengthValue, InvalidType
from tcex.logger.sensitive_filter import SensitiveFilter
from tcex.logger.trace_logger import TraceLogger
from tcex.util.variable import BinaryVariable

# get tcex logger
filter_sensitive = SensitiveFilter(name='sensitive_filter')
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore
_logger.addFilter(filter_sensitive)


class Sensitive:
    """Sensitive Field Type"""

    allow_empty: bool = True
    min_length: int | None = None
    max_length: int | None = None

    def __init__(self, value: str | Self):
        """Initialize the Sensitive object."""
        if isinstance(value, Sensitive):
            self._sensitive_value = value.value
        else:
            self._sensitive_value = value
        filter_sensitive.add(self._sensitive_value)

    @classmethod
    def __get_validators__(cls) -> Generator:
        """Define one or more validators for Pydantic custom type."""
        yield cls.validate_type
        yield cls.validate_allow_empty
        yield cls.validate_max_length
        yield cls.validate_min_length
        yield cls.wrap_type

    def __len__(self) -> int:
        """Return the length of the sensitive value."""
        return len(self._sensitive_value)

    @classmethod
    def __modify_schema__(cls, field_schema: dict[str, Any]):
        """Modify the field schema."""

        def _update_not_none(mapping: dict[Any, Any], **update: Any):
            mapping.update({k: v for k, v in update.items() if v is not None})

        _update_not_none(
            field_schema,
            type='string',
            writeOnly=True,
            format='password',
            minLength=cls.min_length,
            maxLength=cls.max_length,
        )

    def __repr__(self) -> str:
        """."""
        return f"""Sensitive('{self}')"""

    def __str__(self) -> str:
        """Return the value masked.

        If App is running in > DEBUG logging level and the sensitive data is greater
        than X, then show the first and last character of the value. This is very
        helpful in debugging App where the incorrect credential could have been passed.
        """
        # DEBUG or TRACE
        trace_log_level = 10
        if (
            self._sensitive_value
            and _logger.getEffectiveLevel() <= trace_log_level
            and isinstance(self.value, str)
            and len(self.value) >= trace_log_level
        ):
            return f"""{self.value[:1]}{'*' * 4}{self.value[-1:]}"""
        return '**********'

    @classmethod
    def validate_allow_empty(
        cls, value: bytes | str | Self, field: ModelField
    ) -> bytes | str | Self:
        """Raise exception if value is empty and allow_empty is False."""
        if cls.allow_empty is False:
            if isinstance(value, str) and value.replace(' ', '') == '':
                raise InvalidEmptyValue(field_name=field.name)
            if isinstance(value, bytes) and value == b'':
                raise InvalidEmptyValue(field_name=field.name)
            if isinstance(value, Sensitive) and value.value.replace(' ', '') == '':
                raise InvalidEmptyValue(field_name=field.name)
        return value

    @classmethod
    def validate_max_length(
        cls, value: bytes | str | Self, field: ModelField
    ) -> bytes | str | Self:
        """Raise exception if value does not match pattern."""
        if cls.max_length is not None and len(value) > cls.max_length:
            raise InvalidLengthValue(
                field_name=field.name, constraint=cls.max_length, operation='max'
            )
        return value

    @classmethod
    def validate_min_length(
        cls, value: bytes | str | Self, field: ModelField
    ) -> bytes | str | Self:
        """Raise exception if value does not match pattern."""
        if cls.min_length is not None and len(value) < cls.min_length:
            raise InvalidLengthValue(
                field_name=field.name, constraint=cls.min_length, operation='min'
            )
        return value

    @classmethod
    def validate_type(cls, value: bytes | str | Self, field: ModelField) -> bytes | str | Self:
        """Raise exception if value is not a String type."""
        if not isinstance(value, bytes | str | Sensitive):
            raise InvalidType(
                field_name=field.name, expected_types='(bytes, str)', provided_type=str(type(value))
            )
        return value

    @property
    def value(self) -> str:
        """Return the actual value."""
        if not isinstance(self._sensitive_value, BinaryVariable | bytes):
            # file variables can be used for client certs, json credential,
            # etc and the data is provided as a BinaryVariable object. This
            # is a special case where we need to return the value as a string.
            return str(self._sensitive_value)
        return self._sensitive_value

    @classmethod
    def wrap_type(cls, value: str) -> Self:
        """Raise exception if value is not a String type."""
        return cls(value)


def sensitive(
    allow_empty: bool = True,
    max_length: int | None = None,
    min_length: int | None = None,
) -> type[Sensitive]:
    """Return configured instance of String."""
    namespace = {
        'allow_empty': allow_empty,
        'max_length': max_length,
        'min_length': min_length,
    }
    return type('ConstrainedSensitive', (Sensitive,), namespace)



"""TcEx Framework Module"""

# third-party
from requests import Session  # TYPE-CHECKING

# first-party
from tcex.api.tc.ti_transform import TiTransform, TiTransforms
from tcex.api.tc.ti_transform.model import GroupTransformModel, IndicatorTransformModel
from tcex.api.tc.util.threat_intel_util import ThreatIntelUtil
from tcex.api.tc.v2.v2 import V2
from tcex.api.tc.v3.v3 import V3
from tcex.input.input import Input  # TYPE-CHECKING
from tcex.pleb.cached_property import cached_property


class TC:
    """API -> TC

    Args:
        inputs: An instance of the Input class.
        session_tc: An configured instance of request.Session with TC API Auth.
    """

    def __init__(self, inputs: Input, session_tc: Session):
        """Initialize instance properties."""
        self.inputs = inputs
        self.session_tc = session_tc

    @staticmethod
    def group_transform(transform: dict) -> GroupTransformModel:
        """Return a group transform model."""
        return GroupTransformModel(**transform)

    @staticmethod
    def indicator_transform(transform: dict) -> IndicatorTransformModel:
        """Return a indicator transform model."""
        return IndicatorTransformModel(**transform)

    @staticmethod
    def ti_transform(
        ti_dict: dict,
        transforms: list[GroupTransformModel | IndicatorTransformModel],
    ) -> TiTransform:
        """Return an instance of TI Transform class."""
        return TiTransform(ti_dict, transforms)

    @staticmethod
    def ti_transforms(
        ti_dict: list[dict],
        transforms: list[GroupTransformModel | IndicatorTransformModel],
    ) -> TiTransforms:
        """Return an instance of TI Transforms class."""
        return TiTransforms(ti_dict, transforms)

    @cached_property
    def utils(self) -> ThreatIntelUtil:
        """Return an instance of TI Transform class."""
        return ThreatIntelUtil(self.session_tc)

    @cached_property
    def v2(self) -> V2:
        """Return a case management instance."""
        return V2(self.inputs, self.session_tc)

    @cached_property
    def v3(self) -> V3:
        """Return a case management instance."""
        return V3(self.session_tc)



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Field, validator


class PlaybookModel(BaseModel):
    """Playbook Model

    Supported for the following runtimeLevel:
    * Playbook
    """

    tc_playbook_kvstore_context: str = Field(
        None,
        description='The KV Store context for the current App execution.',
        inclusion_reason='runtimeLevel',
    )
    tc_playbook_out_variables: list | None = Field(
        None,
        description='The list of requested output variables.',
        inclusion_reason='runtimeLevel',
    )

    @validator('tc_playbook_out_variables', pre=True)
    @classmethod
    def parse_tc_playbook_out_variables(cls, v):
        """Ensure value is an array."""
        if isinstance(v, str):
            v = v.split(',')
        return v



"""TcEx Framework Module"""

# standard library
from typing import cast

# third-party
import pyaes


class AesOperation:
    """TcEx Utilities AES Operations Class"""

    @staticmethod
    def decrypt_aes_cbc(
        key: bytes | str, ciphertext: bytes | str, iv: bytes | str | None = None
    ) -> bytes:
        """Return AES CBC decrypted string.

        Args:
            key: The encryption key.
            ciphertext: The ciphertext to decrypt.
            iv: The CBC initial vector.
        """
        iv = iv or b'\0' * 16

        # ensure key is bytes
        if isinstance(key, str):
            key = key.encode()

        # ensure iv is bytes
        if isinstance(iv, str):
            iv = iv.encode()

        aes_cbc_decrypt = pyaes.Decrypter(pyaes.AESModeOfOperationCBC(key, iv=iv))
        decrypted = aes_cbc_decrypt.feed(ciphertext)
        decrypted += aes_cbc_decrypt.feed()
        return cast(bytes, decrypted)

    @staticmethod
    def encrypt_aes_cbc(
        key: bytes | str, plaintext: bytes | str, iv: bytes | str | None = None
    ) -> bytes:
        """Return AES CBC encrypted string.

        Args:
            key: The encryption key.
            plaintext: The text to encrypt.
            iv: The CBC initial vector.
        """
        iv = iv or b'\0' * 16

        # ensure key is bytes
        if isinstance(key, str):
            key = key.encode()

        # ensure plaintext is bytes
        if isinstance(plaintext, str):
            plaintext = plaintext.encode()

        # ensure iv is bytes
        if isinstance(iv, str):
            iv = iv.encode()

        aes_cbc_encrypt = pyaes.Encrypter(pyaes.AESModeOfOperationCBC(key, iv=iv))
        encrypted = aes_cbc_encrypt.feed(plaintext)
        encrypted += aes_cbc_encrypt.feed()
        return cast(bytes, encrypted)



"""TcEx Framework Module"""

# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql_type import TqlType


class Tql:
    """ThreatConnect TQL"""

    def __init__(self):
        """Initialize instance properties"""
        self._filters = []
        self.raw_tql = None

    @property
    def as_str(self):
        """Convert the TQL obj to a string"""
        filters = []
        for tql_filter in self.filters:
            # keywords are all one work (e.g. task_id should be taskid)
            keyword = tql_filter['keyword'].replace('_', '')
            value = tql_filter['value']
            try:
                filters.append(f'{keyword}({value._tql.as_str})')  # noqa: SLF001
            except Exception:
                if isinstance(value, list):
                    if tql_filter.get('type') in (TqlType.FLOAT, TqlType.INTEGER):
                        value = [str(int_) for int_ in value]
                    elif tql_filter.get('type') == TqlType.STRING:
                        value = [f'"{str_}"' for str_ in value]
                    value = ','.join(value)
                    value = f'({value})'
                elif tql_filter.get('type') == TqlType.STRING:
                    value = f'"{value}"'
                filters.append(f'{keyword} {tql_filter["operator"].value} {value}')

        return ' and '.join(filters)

    @property
    def filters(self) -> list[dict]:
        """Return the filters"""
        return self._filters

    @filters.setter
    def filters(self, filters: list[dict]):
        """Set the filters"""
        self._filters = filters

    def add_filter(
        self,
        keyword: str,
        operator: Enum | str,
        value: float | list | str | FilterABC | Arrow | datetime,
        type_: TqlType | None = TqlType.STRING,
    ):
        """Add a filter to the current obj

        Args:
            keyword: the field to search on
            operator: the operator to use
            value: the value to compare
            type_: How to treat the value (defaults to String)
        """
        self.filters.append(
            {'keyword': keyword, 'operator': operator, 'value': value, 'type': type_}
        )

    def set_raw_tql(self, tql: str):
        """Set a raw TQL filter"""
        self.raw_tql = tql



"""TcEx Framework Module"""

# standard library
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .tcex_json import TcexJson  # CIRCULAR-IMPORT


class TcexJsonUpdate:
    """Config object for tcex.json configuration file (updater)"""

    def __init__(self, tj: 'TcexJson'):
        """Initialize instance properties."""
        self.tj = tj

    def multiple(self, template: str | None = None):
        """Update the contents of the tcex.json file."""

        # update app_name
        self.update_package_app_name()

        # update deprecated fields
        self.update_deprecated_fields()

        # update package excludes
        self.update_package_excludes()

        # update template
        if template is not None:
            self.tj.model.template_name = template

        # write updated profile
        self.tj.write()

    def update_deprecated_fields(self):
        """Update the lib_versions array in the tcex.json file."""
        if hasattr(self.tj.model, 'lib_versions'):
            self.tj.model.lib_versions = None

    def update_package_app_name(self):
        """Update the package app_name in the tcex.json file."""
        if (
            self.tj.model.package.app_name is None
            or self.tj.model.package.app_name in self.tj.ij.app_prefixes.values()
        ):
            # lower case name and replace prefix if already exists
            _app_name = Path.cwd().name.lower().replace(self.tj.ij.app_prefix.lower(), '')

            # replace spaces and dashes with underscores
            _app_name = _app_name.replace(' ', '_').replace('-', '_').lower()

            # title case app name
            _app_name = '_'.join([a.title() for a in _app_name.split('_')])

            # prepend appropriate App prefix (e.g., TCPB_-_)
            _app_name = f'{self.tj.ij.app_prefix}{_app_name}'

            # update App name
            self.tj.model.package.app_name = _app_name

    def update_package_excludes(self):
        """Update the excludes values in the tcex.json file."""
        for i in [
            '.gitignore',
            '.pre-commit-config.yaml',
            'app_spec.yaml',
            'local-*',
            'pyproject.toml',
        ]:
            if i not in self.tj.model.package.excludes:
                # TODO: [low] pydantic doesn't seem to allow removing items from list???
                self.tj.model.package.excludes.append(i)



"""TcEx Framework Module"""

# standard library
import logging
import os
import platform
import sys
from importlib.metadata import version
from pathlib import Path
from urllib.parse import urlsplit

# third-party
from certifi import where
from requests import Session  # TYPE-CHECKING

# first-party
from tcex.app.config.install_json import InstallJson
from tcex.input.model.common_model import CommonModel  # TYPE-CHECKING
from tcex.logger.api_handler import ApiHandler, ApiHandlerFormatter
from tcex.logger.cache_handler import CacheHandler
from tcex.logger.rotating_file_handler_custom import RotatingFileHandlerCustom
from tcex.logger.trace_logger import TraceLogger


class Logger:
    """Framework logger module."""

    def __init__(self, logger_name: str):
        """Initialize instance properties."""
        self.logger_name = logger_name

        # properties
        self.ij = InstallJson()

    @property
    def _logger(self) -> TraceLogger:
        """Return the logger. The inputs.model property is not available in init."""
        logging.setLoggerClass(TraceLogger)
        logger = logging.getLogger(self.logger_name)
        logger.setLevel(logging.TRACE)  # type: ignore
        return logger  # type: ignore

    @property
    def _formatter(self) -> logging.Formatter:
        """Return log formatter."""
        tx_format = (
            '%(asctime)s - %(name)s - %(levelname)8s - %(message)s '
            '(%(filename)s:%(funcName)s:%(lineno)d)'
        )
        return logging.Formatter(tx_format)

    @property
    def _formatter_thread_name(self) -> logging.Formatter:
        """Return log formatter."""
        tx_format = (
            '%(asctime)s - %(name)s - %(levelname)8s - %(message)s '
            '(%(filename)s:%(funcName)s:%(lineno)d:%(threadName)s)'
        )
        return logging.Formatter(tx_format)

    def handler_exist(self, handler_name: str) -> bool:
        """Remove a file handler by name.

        Args:
            handler_name: The handler name to remove.

        Returns:
            bool: True if handler current exists
        """
        return any(h.get_name() == handler_name for h in self._logger.handlers)

    @property
    def log(self) -> TraceLogger:
        """Return logger."""
        return self._logger

    @staticmethod
    def log_level(level: str | None) -> int:
        """Return proper level from string.

        Args:
            level: The logging level.

        Returns:
            int: The logging level as an int.
        """
        level = level or 'debug'
        return logging.getLevelName(level.upper())

    def remove_handler_by_name(self, handler_name: str):
        """Remove a file handler by name.

        Args:
            handler_name: The handler name to remove.
        """
        for h in self._logger.handlers:
            if h.get_name() == handler_name:
                self._logger.removeHandler(h)
                break

    def replay_cached_events(self, handler_name: str = 'cache'):
        """Replay cached log events and remove handler."""
        for h in self._logger.handlers:
            if h.get_name() == handler_name:
                events = h.events  # type: ignore
                self._logger.removeHandler(h)
                for event in events:
                    self._logger.handle(event)
                break

        # remove the cache handler
        self.remove_handler_by_name(handler_name=handler_name)

    def shutdown(self):
        """Close all handlers.

        Args:
            handler_name (str): The handler name to remove.
        """
        for h in self._logger.handlers:
            self._logger.removeHandler(h)

    def update_handler_level(self, level: str):
        """Update all handlers log level.

        Args:
            level: The logging level.
        """
        level_ = self.log_level(level)

        # update all handler logging levels
        for h in self._logger.handlers:
            h.setLevel(level_)

    #
    # handlers
    #

    def add_api_handler(self, session_tc: Session, name: str = 'api', level: str | None = None):
        """Add API logging handler.

        Args:
            session_tc: An configured instance of request.Session with TC API Auth.
            name: The name of the handler.
            level: The level value as a string.
        """
        self.remove_handler_by_name(name)
        api = ApiHandler(session_tc)
        api.set_name(name)
        api.setLevel(self.log_level(level))
        api.setFormatter(ApiHandlerFormatter())
        self._logger.addHandler(api)

    def add_cache_handler(self, name: str):
        """Add cache logging handler.

        Args:
            name: The name of the handler.
        """
        self.remove_handler_by_name(name)
        cache = CacheHandler()
        cache.set_name(name)
        # set logging level to INFO as event will typically
        # be only those that happen before args are processed
        cache.setLevel(self.log_level('trace'))
        cache.setFormatter(self._formatter)
        self._logger.addHandler(cache)

    def add_rotating_file_handler(
        self,
        name: str,
        filename: str,
        path: Path | str,
        backup_count: int,
        max_bytes: int,
        level: str,
        formatter: logging.Formatter | None = None,
        mode: str = 'a',
    ):
        """Add custom file logging handler.

        Args:
            name: The name of the handler.
            filename: The name of the logfile.
            path: The path for the logfile.
            backup_count: The maximum # of backup files.
            max_bytes: The max file size before rotating.
            level: The logging level.
            formatter: The logging formatter to use.
            mode: The write mode for the file.
        """
        self.remove_handler_by_name(name)
        formatter = formatter or self._formatter_thread_name

        # create customized handler
        fh = RotatingFileHandlerCustom(
            str(Path(path) / filename), backupCount=backup_count, maxBytes=max_bytes, mode=mode
        )
        fh.set_name(name)
        fh.setFormatter(formatter)
        fh.setLevel(self.log_level(level))
        self._logger.addHandler(fh)

    def add_stream_handler(
        self,
        name: str = 'sh',
        formatter: logging.Formatter | None = None,
        level: str | None = None,
    ):
        """Add stream logging handler.

        Args:
            name: The name of the handler.
            formatter: The logging formatter to use.
            level: The logging level.
        """
        self.remove_handler_by_name(name)
        formatter = formatter or self._formatter
        # create handler
        sh = logging.StreamHandler()
        sh.set_name(name)
        sh.setFormatter(formatter)
        sh.setLevel(self.log_level(level))
        self._logger.addHandler(sh)

    #
    # App info logging
    #

    def log_info(self, inputs: CommonModel):
        """Send System and App data to logs.

        Args:
            inputs: The inputs model.
        """
        self._log_app_data()
        self._log_platform()
        self._log_python_version()
        self._log_environment()
        self._log_tcex_version()
        self._log_tc_proxy(inputs)

    def _log_app_data(self):
        """Log the App data information as a best effort."""
        try:
            self.log.info(f'app-name="{self.ij.model.display_name}"')
            if self.ij.model.app_id is not None:
                self.log.info(f'app-id={self.ij.model.app_id}')
            if self.ij.model.features:
                self.log.info(f"""app-features={','.join(self.ij.model.features)}""")
            self.log.info(f'app-minimum-threatconnect-version={self.ij.model.min_server_version}')
            self.log.info(f'app-runtime-level={self.ij.model.runtime_level}')
            app_version = f'app-version={self.ij.model.program_version}'
            if self.ij.model.commit_hash is not None:
                app_version += f', app-commit-hash={self.ij.model.commit_hash}'
            self.log.info(app_version)  # version and commit hash
        except Exception:  # nosec; pragma: no cover
            pass

    def _log_environment(self):
        """Log the current environment variables."""
        self.log.info(f"""env-all-proxy={self._sanitize_proxy_url(os.getenv('ALL_PROXY'))}""")
        self.log.info(f"""env-http-proxy={self._sanitize_proxy_url(os.getenv('HTTP_PROXY'))}""")
        self.log.info(f"""env-https-proxy={self._sanitize_proxy_url(os.getenv('HTTPS_PROXY'))}""")
        self.log.info(f"""env-no-proxy={os.getenv('NO_PROXY')}""")
        self.log.info(f"""env-curl-ca-bundle={os.getenv('CURL_CA_BUNDLE')}""")
        self.log.info(f"""env-request-ca-bundle={os.getenv('REQUESTS_CA_BUNDLE')}""")
        self.log.info(f"""env-ssl-cert-file={os.getenv('SSL_CERT_FILE')}""")

        # log certifi where
        self.log.info(f"""certifi-where={where()}""")

    def _log_platform(self):
        """Log the current Platform."""
        self.log.info(f'platform="{platform.platform()}"')
        self.log.info(f'pid={os.getpid()}')

    def _log_python_version(self):
        """Log the current Python version."""
        self.log.info(
            f'python-version={sys.version_info.major}.'
            f'{sys.version_info.minor}.'
            f'{sys.version_info.micro}'
        )

    def _log_tc_proxy(self, inputs: CommonModel):
        """Log the proxy settings.

        Args:
            inputs: The inputs model.
        """
        if inputs.tc_proxy_tc:
            self.log.info(f'proxy-server-tc={inputs.tc_proxy_host}:{inputs.tc_proxy_port}')

    def _log_tcex_version(self):
        """Log the current TcEx version number."""
        try:
            self.log.info(f"""tcex-version={version('tcex')}""")
        except ImportError:
            self.log.warning('TcEx version no found.')

    @staticmethod
    def _sanitize_proxy_url(url: str | None) -> str | None:
        """Sanitize a proxy url, masking username and password."""
        if url is not None:
            url_data = urlsplit(url)
            if all([url_data.username, url_data.password]):
                url = url.replace(f'{url_data.username}:{url_data.password}', '***:***')
        return url



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Field


class OrganizationModel(BaseModel):
    """Organization Model

    Supported for the following runtimeLevel:
    * Organization
    """

    tc_job_id: int | None = Field(
        None,
        description='The Job Id for the current App execution.',
        inclusion_reason='runtimeLevel',
    )



"""TcEx Framework Module"""

# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class ResultModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='Result Model',
    validate_assignment=True,
):
    """Result Model"""

    _associated_type = PrivateAttr(default=False)
    _cm_type = PrivateAttr(default=False)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    archived: bool = Field(
        None,
        description='Has the result been archived?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='archived',
    )
    archived_date: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the Entity was archived.',
        read_only=True,
        title='archivedDate',
    )
    associated: bool = Field(
        None,
        description='Has the result been associated to an entity within Threatconnect?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associated',
    )
    false_positive: bool = Field(
        None,
        description='Is the result declared false positive?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='falsePositive',
    )
    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    intel_req_id: int | None = Field(
        None,
        allow_mutation=False,
        description='The id of the intel requirement that the result is associated.',
        read_only=True,
        title='intelReqId',
    )
    intel_requirement: dict | None = Field(
        None,
        allow_mutation=False,
        description='The intel requirement associated to the result.',
        read_only=True,
        title='intelRequirement',
    )
    internal: bool = Field(
        None,
        allow_mutation=False,
        description='Is the result sourced internally from Threatconnect.',
        read_only=True,
        title='internal',
    )
    item_id: int | None = Field(
        None,
        allow_mutation=False,
        description='The id of the entity that matched the result.',
        read_only=True,
        title='itemId',
    )
    item_type: str | None = Field(
        None,
        allow_mutation=False,
        description='The type of the entity that matched the result.',
        read_only=True,
        title='itemType',
    )
    matched_date: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the result last matched with the intel requirement.',
        read_only=True,
        title='matchedDate',
    )
    name: str | None = Field(
        None,
        allow_mutation=False,
        description='The name of the result.',
        read_only=True,
        title='name',
    )
    origin: str | None = Field(
        None,
        allow_mutation=False,
        description='The origin of the result if derived from an internal or external source.',
        read_only=True,
        title='origin',
    )
    owner_id: int | None = Field(
        None,
        allow_mutation=False,
        description='The organization id that the result belongs.',
        read_only=True,
        title='ownerId',
    )
    owner_name: str | None = Field(
        None,
        allow_mutation=False,
        description='The organization name that the result belongs.',
        read_only=True,
        title='ownerName',
    )
    score: int | None = Field(
        None,
        allow_mutation=False,
        description='The relevancy score.',
        read_only=True,
        title='score',
    )


class ResultDataModel(
    BaseModel,
    title='Result Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Results Data Model"""

    data: list[ResultModel] | None = Field(
        [],
        description='The data for the Results.',
        methods=['POST', 'PUT'],
        title='data',
    )


class ResultsModel(
    BaseModel,
    title='Results Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Results Model"""

    _mode_support = PrivateAttr(default=False)

    data: list[ResultModel] | None = Field(
        [],
        description='The data for the Results.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# add forward references
ResultDataModel.update_forward_refs()
ResultModel.update_forward_refs()
ResultsModel.update_forward_refs()



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Extra, Field, validator

# first-party
from tcex.pleb.cached_property import cached_property
from tcex.util import Util
from tcex.util.render.render import Render
from tcex.util.string_operation import CamelString


class ExtraModel(
    BaseModel,
    extra=Extra.forbid,
    title='Extra Model',
    validate_assignment=True,
):
    """Model Definition"""

    alias: CamelString = Field(..., description='Field alias.')
    import_data: str | None = Field(None, description='The import data.')
    import_source: str | None = Field(
        None, description='The source of the import: standard library, first-party, etc.'
    )
    methods: list[str] = Field([], description='Field methods.')
    model: str | None = Field(None, description='The type model.')
    type: CamelString = Field(..., description='The type of the property.')
    typing_type: str = Field(..., description='The Python typing hint type.')

    @validator('alias', 'type', always=True, pre=True)
    @classmethod
    def _camel_string(cls, v):
        """Convert to CamelString."""
        return CamelString(v)


class PropertyModel(
    BaseModel,
    alias_generator=Util().snake_to_camel,
    arbitrary_types_allowed=True,
    extra=Extra.forbid,
    keep_untouched=(cached_property,),
    title='Property Model',
    validate_assignment=True,
):
    """Model Definition"""

    applies_to: list[str] | None = Field(None, description='Applies to property.')
    allowable_values: list[str] | None = Field(
        None, description='Allowable values for the property.'
    )
    common: bool = Field(default=False, description='Common property.')
    conditional_read_only: list[str] | None = Field(
        None, description='Conditionally read-only property.'
    )
    conditional_required: list[str] | None = Field(
        None, description='Conditionally required property.'
    )
    default: str | None = Field(default=None, description='Default value of the property value.')
    description: str | None = Field(default=None, description='Description of the property.')
    max_size: int | None = Field(default=None, description='Maximum size of the property value.')
    max_length: int | None = Field(
        default=None, description='Maximum length of the property value.'
    )
    max_value: int | None = Field(default=None, description='Maximum value of the property value.')
    min_length: int | None = Field(
        default=None, description='Minimum length of the property value.'
    )
    min_value: int | None = Field(default=None, description='Minimum value of the property value.')
    name: CamelString = Field(default=None, description='Name of the property.')
    read_only: bool = Field(default=False, description='Read only property.')
    required: bool = Field(default=False, description='Required property.')
    required_alt_field: str | None = Field(None, description='Required alternative field property.')
    type: CamelString = Field(..., description='The defined property type.')
    updatable: bool = Field(default=True, description='Updatable property.')

    @validator('name', 'type', always=True, pre=True)
    @classmethod
    def _came_string(cls, v):
        """Convert to CamelString."""
        if v is None:
            ex_msg = 'The property must have a name.'
            raise ValueError(ex_msg)
        return CamelString(v)

    @validator('description', always=True, pre=True)
    @classmethod
    def _fix_first_letter_of_sentence(cls, v):
        """Convert to CamelString."""
        if v:
            # ensure the first letter of the sentence is capitalized
            v = v[0].capitalize() + v[1:]
        return v

    @cached_property
    def extra(self) -> ExtraModel:
        """Return the extra model."""
        extra_data = self.__extra_data()
        # print(self.name, self.type, extra_data)
        return ExtraModel(**extra_data)  # type: ignore

    def __extra_data(self) -> dict[str, str]:
        """Enrich the extra model."""
        # set default value for type
        extra = {
            'alias': self.name,
            # calculate the field methods
            'methods': self.__calculate_methods(),
            # set type to current type value, update later
            'type': self.type,
        }

        reserved_words = ['from']
        if self.name in reserved_words:
            extra['alias'] = f'{self.name}_'

        # process bool types
        self.__process_bool_types(self, extra)

        # process dict types
        self.__process_dict_types(self, extra)

        # process float types
        self.__process_float_types(self, extra)

        # process int types
        self.__process_int_types(self, extra)

        # process str types
        self.__process_str_types(self, extra)

        # process tc types
        self.__process_tc_types(self, extra)

        # process special types
        self.__process_special_types(self, extra)

        if extra.get('typing_type') is None:
            Render.panel.failure(
                f'New type found for object name="{self.name}"\n'
                f'Type: "{self.type}" should be added to PropertyModel '
                f'in one of the following methods:\n'
                f'_process_dict, _process_tc_type, _process_special_types\n'
                f'\nextra: {extra}'
            )

        return extra

    def __calculate_methods(self):
        """Calculate the field methods."""

        # special case until core updates OPTIONS output
        methods = []
        if self.name.snake_case() == 'deleted_reason':
            methods.append('DELETE')
        elif self.read_only is False:
            methods.append('POST')

            if self.updatable is not False:
                methods.append('PUT')
        return methods

    @classmethod
    def __process_bool_types(cls, pm: 'PropertyModel', extra: dict[str, str]):
        """Process standard type."""
        types = ['boolean']
        if pm.type.lower() in types:
            extra.update({'typing_type': 'bool'})

    @classmethod
    def __process_dict_types(cls, pm: 'PropertyModel', extra: dict[str, str]):
        """Process standard type."""
        types = [
            'AiInsights',
            'AttackSecurityCoverage',
            'AttributeSource',
            'DNSResolutions',
            'Enrichments',
            'GeoLocation',
            'InvestigationLinks',
            'IntelligenceReviews',
            'IntelReqType',
            'IntelRequirement',
            'KeywordSection',
            'Links',
            'Map',
            'ValidationRule',
            'Strings',
            'WhoIs',
            'Exclusions',
        ]
        if pm.type in types:
            extra.update({'typing_type': cls.__extra_format_type('dict')})

    @classmethod
    def __process_float_types(cls, pm: 'PropertyModel', extra: dict[str, str]):
        """Process standard type."""
        types = ['Double']
        if pm.type in types:
            extra.update({'typing_type': cls.__extra_format_type('float')})

    @classmethod
    def __process_int_types(cls, pm: 'PropertyModel', extra: dict[str, str]):
        """Process standard type."""
        types = [
            'bigdecimal',  # BigDecimal
            'integer',  # Integer
            'long',  # Long
            'short',  # Short
        ]
        if pm.type.lower() in types:
            extra.update({'typing_type': cls.__extra_format_type('int')})

    @classmethod
    def __process_special_types(cls, pm: 'PropertyModel', extra: dict[str, str]):
        """Process standard type."""
        bi = 'from tcex.api.tc.v3.'

        if pm.type == 'Assignee':
            bi += 'security.assignee_model'
            extra.update(
                {
                    'import_data': f'{bi} import AssigneeModel',
                    'import_source': 'first-party-forward-reference',
                    'model': f'{pm.type}Model',
                    'typing_type': cls.__extra_format_type_model(pm.type),
                }
            )
        elif pm.type == 'AttributeDatas':
            extra.update(
                {
                    'import_data': f'{bi}.attributes.attribute_model import AttributesModel',
                    'import_source': 'first-party-forward-reference',
                    'model': f'{pm.type}Model',
                    'typing_type': cls.__extra_format_type_model(pm.type),
                }
            )
        elif pm.type == 'CaseAttributes':
            bi += 'case_attributes.case_attribute_model'
            extra.update(
                {
                    'import_data': f'{bi} import CaseAttributesModel',
                    'import_source': 'first-party-forward-reference',
                    'model': f'{pm.type}Model',
                    'typing_type': cls.__extra_format_type_model(pm.type),
                }
            )
        elif pm.type == 'Date':
            extra.update(
                {
                    'import_data': 'from datetime import datetime',
                    'import_source': 'standard library',
                    'typing_type': cls.__extra_format_type('datetime'),
                }
            )
        elif pm.type == 'Group':
            bi += 'groups.group_model'
            extra.update(
                {
                    'import_data': f'{bi} import GroupModel',
                    'import_source': 'first-party-forward-reference',
                    'model': f'{pm.type}Model',
                    'typing_type': cls.__extra_format_type_model(pm.type),
                }
            )
        elif pm.type == 'GroupAttributes':
            bi += 'group_attributes.group_attribute_model'
            extra.update(
                {
                    'import_data': f'{bi} import GroupAttributesModel',
                    'import_source': 'first-party-forward-reference',
                    'model': f'{pm.type}Model',
                    'typing_type': cls.__extra_format_type_model(pm.type),
                }
            )
        elif pm.type == 'Indicator':
            bi += 'indicators.indicator_model'
            extra.update(
                {
                    'import_data': f'{bi} import IndicatorModel',
                    'import_source': 'first-party-forward-reference',
                    'model': f'{pm.type}Model',
                    'typing_type': cls.__extra_format_type_model(pm.type),
                }
            )
        elif pm.type == 'IntelReqType':
            bi += 'intel_requirements.intel_req_type_model'
            extra.update(
                {
                    'import_data': f'{bi} import IntelReqTypeModel',
                    'import_source': 'first-party-forward-reference',
                    'model': 'IntelReqTypeModel',
                    'typing_type': cls.__extra_format_type_model(pm.type),
                }
            )
        elif pm.type == 'IndicatorAttributes':
            bi += 'indicator_attributes.indicator_attribute_model'
            extra.update(
                {
                    'import_data': f'{bi} import IndicatorAttributesModel',
                    'import_source': 'first-party-forward-reference',
                    'model': f'{pm.type}Model',
                    'typing_type': cls.__extra_format_type_model(pm.type),
                }
            )
        elif pm.type == 'JsonNode':
            extra.update(
                {
                    'import_source': 'standard library',
                    'typing_type': """dict | list[dict] | None""",
                }
            )
        elif pm.type == 'KeywordSection':
            bi += 'intel_requirements.keyword_sections.keyword_section_model'
            extra.update(
                {
                    'import_data': f'{bi} import KeywordSectionModel',
                    'import_source': 'first-party-forward-reference',
                    'model': 'KeywordSectionModel',
                    'typing_type': f'list[{cls.__extra_format_type_model(pm.type)}]',
                }
            )
        elif pm.name == 'customAssociationNames':
            extra.update(
                {
                    'typing_type': 'list[str]',
                }
            )
        elif pm.type == 'TaskAssignees':
            extra.update(
                {
                    'import_data': (f'{bi}security.task_assignee_model import TaskAssigneesModel'),
                    'import_source': 'first-party-forward-reference',
                    'model': f'{pm.type}Model',
                    'typing_type': cls.__extra_format_type_model(pm.type),
                }
            )
        elif pm.type == 'VictimAttributes':
            bi += 'victim_attributes.victim_attribute_model'
            extra.update(
                {
                    'import_data': f'{bi} import VictimAttributesModel',
                    'import_source': 'first-party-forward-reference',
                    'model': f'{pm.type}Model',
                    'typing_type': cls.__extra_format_type_model(pm.type),
                }
            )

    @classmethod
    def __process_str_types(cls, pm: 'PropertyModel', extra: dict[str, str]):
        """Process standard type."""
        types = [
            'From',
            'Owner',
            'String',
        ]
        if pm.type in types:
            extra.update({'typing_type': cls.__extra_format_type('str')})

    @classmethod
    def __process_tc_types(cls, pm: 'PropertyModel', extra: dict[str, str]):
        """Process standard type."""
        types = [
            'AdversaryAssets',
            'Artifact',
            'Artifacts',
            'ArtifactType',
            'Case',
            'Cases',
            'FileActions',
            'FileOccurrences',
            'Groups',
            'Indicators',
            'Note',
            'Notes',
            'SecurityLabel',
            'SecurityLabels',
            'Tag',
            'Tags',
            'Task',
            'Tasks',
            'User',
            'Users',
            'Victims',
            'VictimAssets',
            'WorkflowEvent',
            'WorkflowEvents',
            'WorkflowTemplate',
        ]
        if pm.type in types:
            extra.update(cls.__extra_gen_req_code(pm.type))
            extra['model'] = f'{pm.type}Model'
            extra['typing_type'] = cls.__extra_format_type_model(pm.type)

    @classmethod
    def __extra_gen_req_code(cls, type_: CamelString) -> dict[str, str]:
        """Return the requirements code"""
        type_ = Util().camel_string(type_)
        return {
            'import_data': (
                f'from {cls.__extra_tap(type_)}.{type_.plural().snake_case()}.'
                f'{type_.singular().snake_case()}_model import {type_}Model'
            ),
            'import_source': 'first-party-forward-reference',
        }

    @classmethod
    def __extra_format_type(cls, type_: str, optional: bool = True) -> str:
        """Format type for use in code."""
        if optional:
            return f'{type_} | None'
        return type_

    @classmethod
    def __extra_format_type_non_native(cls, type_: str, optional: bool = False) -> str:
        """Format type for use in code."""
        if optional:
            return f"'{type_} | None'"
        return f"'{type_}'"

    @classmethod
    def __extra_format_type_model(cls, type_: str, optional: bool = False) -> str:
        """Format type for use in code."""
        return cls.__extra_format_type_non_native(f'{type_}Model', optional)

    @classmethod
    def __extra_tap(cls, type_: CamelString) -> str:
        """Return the TcEx Api Path."""
        if type_.snake_case().plural().lower() in [
            'owners',
            'owner_roles',
            'system_roles',
            'users',
            'user_groups',
            'exclusion_lists',
        ]:
            return 'tcex.api.tc.v3.security'

        if type_.snake_case().plural().lower() in [
            'categories',
            'results',
            'subtypes',
        ]:
            return 'tcex.api.tc.v3.intel_requirements'

        return 'tcex.api.tc.v3'



"""TcEx Framework Module"""

# standard library
from collections.abc import Generator
from typing import Any

# third-party
from pydantic import ValidationError
from requests.exceptions import ProxyError

# first-party
from tcex.api.tc.v3._gen._options_abc import OptionsABC
from tcex.api.tc.v3._gen.model._filter_model import FilterModel
from tcex.pleb.cached_property import cached_property
from tcex.util.render.render import Render

# The OptionsData class is used to download the options data for a given object type and
# correct and/or update the data to ensure it is usable for the code generation process.
# The "plan" is to have a single class that handles downloading the data and updating it.
# The contents* methods download, update, and load to the model. The content_models method
# is typically the entry point to this class.


class OptionsDataFilter(OptionsABC):
    """Gen Model Download Options Class"""

    def __init__(self, api_url: str, object_type: str):
        """Initialize instance properties."""
        self.api_url = api_url
        self.object_type = object_type

        # properties
        self.messages = []

    @cached_property
    def contents(self) -> list[dict[str, Any]]:
        """Return defined API properties for the current object.

        Response:
        {
            "keyword": "analyticsScore",
            "name": "Analytics Score",
            "type": "Integer",
            "description": "The intel score of the artifact",
            "groupable": false,
            "targetable": true
        }
        """
        _properties = []
        try:
            r = self.session.options(f'{self.api_url}/tql', params={})
            # print(r.request.method, r.request.url, r.text)
            if r.ok:
                _properties = r.json().get('data', [])
        except (ConnectionError, ProxyError) as ex:
            Render.panel.failure(f'Failed getting types properties ({ex}).')

        return _properties

    @property
    def contents_updated(self) -> Generator:
        """Update the properties contents, fixing issues in core data."""
        # Provided invalid type "BigInteger" on tql options call and correcting it to String
        for filter_data in self.contents:
            if self.object_type == 'indicators':
                title = 'Indicator Keyword Type'
                if filter_data['keyword'] == 'addressIpval' and filter_data['type'] != 'String':
                    self.messages.append(f'- [{self.object_type}] - ({title}) - fix required.')
                    filter_data['type'] = 'String'
                else:
                    self.messages.append(f'- [{self.object_type}] - ({title}) - fix NOT required.')

            if self.object_type == 'cases' and 'description' in filter_data:
                # fix misspelling in core data
                miss_map = {
                    'occured': 'occurred',
                    'Threatassess': 'ThreatAssess',
                }
                for c, w in miss_map.items():
                    if c in filter_data['description']:
                        filter_data['description'] = filter_data['description'].replace(c, w)

            yield filter_data

    @cached_property
    def content_models(self) -> Generator[FilterModel, None, None]:
        for field_data in self.contents_updated:
            try:
                yield FilterModel(**field_data)
            except ValidationError as ex:
                Render.panel.failure(f'Failed generating filter model: data={field_data} ({ex}).')



"""TcEx Framework Module"""

# standard library
import logging

# third-party
from pydantic import BaseModel

# first-party
from tcex.logger.trace_logger import TraceLogger
from tcex.pleb.cached_property import cached_property

_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class NAICSTag(BaseModel):
    """NAICTag Class"""

    id: str
    name: str

    @property
    def formatted(self):
        """Return the formatted tag."""
        return f'NAICS: {self.id} - {self.name}'


class NAICSTags:
    """NAICTags Class"""

    def __init__(self, naics_tags: dict[str, str], verbose: bool = False):
        """Initialize instance properties."""
        self._naics_tags = {id_: NAICSTag(id=id_, name=name) for id_, name in naics_tags.items()}
        self.verbose = verbose
        self.log = _logger

    @cached_property
    def naics_tags_name_id(self) -> dict[str, list[NAICSTag]]:
        """Return a dict of MitreTags keyed by name."""
        naics_tags = {}
        for tag in self._naics_tags.values():
            naics_tags.setdefault(tag.name.lower(), []).append(NAICSTag(id=tag.id, name=tag.name))

        return naics_tags

    def get_by_name(self, name: str, default: list[str] | None = None) -> list[str] | None:
        """Return the tag id for the provided name."""
        naics_tags = self.naics_tags_name_id.get(name.lower(), [])
        tags = []
        for tag in naics_tags:
            tags.append(tag.formatted)
        if not tags and self.verbose:
            self.log.warning(f'No NAICS match found for {name}.')
        return tags or default

    def get_by_id(self, id_: str | int, default=None) -> str | None:
        """Return the tag name for the provided id."""
        if isinstance(id_, int):
            id_ = str(id_)
        tag = self._naics_tags.get(id_)
        if not tag:
            if self.verbose is True:
                self.log.warning(f'No NAICS match found for {id_}.')
            return default
        return tag.formatted

    def get_all_by_id(self, id_: str | int, default=None) -> list[str] | None:
        """Return the tag name for the provided id."""
        id_ = str(id_)

        if not self._naics_tags.get(id_):
            if self.verbose:
                self.log.warning(f'No NAICS match found for {id_}.')
            return default

        parent = id_[:2]
        tags = []
        tag = self._naics_tags.get(parent)
        if tag:
            tags.append(tag.formatted)

        remainder = id_[2:]
        if not remainder:
            if not tags and self.verbose:
                self.log.warning(f'No NAICS match found for {id_}.')
            return tags or default

        current_id = parent
        for subtype in remainder:
            current_id += subtype
            if tag := self._naics_tags.get(current_id):
                tags.append(tag.formatted)
        if not tags and self.verbose:
            self.log.warning(f'No NAICS match found for {id_}.')
        return tags or default



"""TcEx Framework Module"""

# standard library
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc.v2.threat_intelligence.mapping.mapping import Mapping

if TYPE_CHECKING:
    # first-party
    from tcex.api.tc.v2.threat_intelligence.threat_intelligence import ThreatIntelligence

# import local modules for dynamic reference
module = __import__(__name__)


class SecurityLabel(Mapping):
    """Unique API calls for SecurityLabel API Endpoints"""

    def __init__(self, ti: 'ThreatIntelligence', name, **kwargs):
        """."""
        super().__init__(
            ti,
            main_type='SecurityLabel',
            api_type='securitylabels',
            sub_type=None,
            api_entity='securityLabel',
            api_branch=None,
            owner=kwargs.pop('owner', None),
        )
        self._data['type'] = 'securityLabels'
        self._data['sub_type'] = None
        self._data['name'] = name

        for arg, value in kwargs.items():
            self.add_key_value(arg, value)

    @property
    def as_entity(self):
        """Return the object as an entity."""
        return {}

    @staticmethod
    def is_security_label():
        """Return indication that this is a security label object."""
        return True

    def can_create(self):
        """Return true if security label can be created."""
        return bool(self._data.get('name'))

    def add_key_value(self, key, value):
        """Convert the value and adds it as a data field."""
        self._data[key] = value

    def name(self, name):
        """Update the security labels name."""
        self._data['name'] = name
        data = {'name': name}
        return self._tc_requests.update(
            self.api_type, self.api_branch, self.unique_id, data, owner=self.owner
        )

    def color(self, color):
        """Update the security labels color."""
        self._data['color'] = color
        data = {'color': color}
        return self._tc_requests.update(
            self.api_type, self.api_branch, self.unique_id, data, owner=self.owner
        )

    def description(self, description):
        """Update the security labels description."""
        self._data['description'] = description
        data = {'description': description}
        return self._tc_requests.update(
            self.api_type, self.api_branch, self.unique_id, data, owner=self.owner
        )

    def date_added(self, date_added):
        """Update the security labels date_added.

        Args:
            date_added: Converted to %Y-%m-%dT%H:%M:%SZ date format
        """
        date_added = self.util.any_to_datetime(date_added).strftime('%Y-%m-%dT%H:%M:%SZ')

        self._data['dateAdded'] = date_added
        data = {'dateAdded': date_added}
        return self._tc_requests.update(
            self.api_type, self.api_branch, self.unique_id, data, owner=self.owner
        )



"""TcEx Framework Module"""

# standard library
import logging
import re

# third-party
from pydantic import BaseModel

# first-party
from tcex.logger.trace_logger import TraceLogger
from tcex.pleb.cached_property import cached_property

_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class MitreTag(BaseModel):
    """MitreTag Class"""

    id: str
    name: str

    @property
    def formatted(self):
        """Return the formatted tag."""
        return f'{self.id} - {self.name}'


class MitreTags:
    """MitreTags Class"""

    def __init__(self, mitre_tags: dict[str, str], verbose: bool = False):
        """Initialize instance properties."""
        self._mitre_tags = {id_: MitreTag(id=id_, name=name) for id_, name in mitre_tags.items()}
        self.verbose = verbose
        self.log = _logger

    @cached_property
    def mitre_tags_name_id(self) -> dict[str, MitreTag]:
        """Return a dict of MitreTags keyed by name."""
        mitre_tags = {}
        for tag in self._mitre_tags.values():
            titles = tag.name.split(': ')

            key = titles[1].strip() if len(titles) > 1 else tag.name
            mitre_tags[key.lower()] = MitreTag(id=tag.id, name=tag.name)

        return mitre_tags

    def get_by_name(self, name: str, default: str | None = None) -> str | None:
        """Return the tag id for the provided name."""
        mitre_tag = self.mitre_tags_name_id.get(name.lower())
        if mitre_tag is None:
            if self.verbose is True:
                self.log.warning(f'No Mitre match found for {name}.')
            return default
        return mitre_tag.formatted

    def get_by_id(self, id_: str, default: str | None = None) -> str | None:
        """Return the tag name for the provided id (e.g., T1000)."""
        mitre_tag = self._mitre_tags.get(str(id_).upper())
        if mitre_tag is None:
            if self.verbose is True:
                self.log.warning(f'No Mitre match found for {id_}, returning id unformatted.')
            return default
        return mitre_tag.formatted

    def get_by_id_regex(self, value: str, default: str | None = None) -> str | None:
        r"""Get the appropriate MitreTag using the (T\d+(?:\.\d+)?) regex."""
        matches = re.findall(r'([Tt]\d+(?:\.\d+)?)', value)
        if not matches:
            if self.verbose is True:
                self.log.warning(f'No Mitre matches found for {value}')
            return default
        if len(matches) > 1:
            if self.verbose is True:
                self.log.warning(f'Multiple Mitre matches found for {value}: {matches}')
            return default
        return self.get_by_id(matches[0], default)



"""TcEx Framework Module"""

# standard library
from collections.abc import Iterator

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.security.exclusion_lists.exclusion_list_filter import ExclusionListFilter
from tcex.api.tc.v3.security.exclusion_lists.exclusion_list_model import (
    ExclusionListModel,
    ExclusionListsModel,
)


class ExclusionList(ObjectABC):
    """ExclusionLists Object.

    Args:
        active (bool, kwargs): Whether the rule is active or not.
        fixed_values (object, kwargs): A list of exclusions represent fixed values.
        managed (bool, kwargs): Whether the rule is managed by the system.
        name (str, kwargs): The name of the rule.
        owner (str, kwargs): The name of the Owner of the Exclusion List.
        owner_id (int, kwargs): The ID of the owner of this exclusion list.
        variable_values (object, kwargs): A list of exclusions serve as wildcards.
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model: ExclusionListModel = ExclusionListModel(**kwargs)
        self._nested_field_name = 'exclusionLists'
        self._nested_filter = 'has_exclusion_list'
        self.type_ = 'Exclusion List'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.EXCLUSION_LISTS.value

    @property
    def model(self) -> ExclusionListModel:
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: dict | ExclusionListModel):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            ex_msg = f'Invalid data type: {type(data)} provided.'
            raise RuntimeError(ex_msg)  # noqa: TRY004


class ExclusionLists(ObjectCollectionABC):
    """ExclusionLists Collection.

    # Example of params input
    {
        "result_limit": 100,  # Limit the retrieved results.
        "result_start": 10,  # Starting count used for pagination.
        "fields": ["caseId", "summary"]  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = ExclusionListsModel(**kwargs)
        self.type_ = 'exclusion_lists'

    def __iter__(self) -> Iterator[ExclusionList]:
        """Return CM objects."""
        return self.iterate(base_class=ExclusionList)  # type: ignore

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.EXCLUSION_LISTS.value

    @property
    def filter(self) -> ExclusionListFilter:
        """Return the type specific filter object."""
        return ExclusionListFilter(self.tql)



"""TcEx Framework Module"""

# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class TagModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='Tag Model',
    validate_assignment=True,
):
    """Tag Model"""

    _associated_type = PrivateAttr(default=False)
    _cm_type = PrivateAttr(default=False)
    _shared_type = PrivateAttr(default=True)
    _staged = PrivateAttr(default=False)

    cases: 'CasesModel' = Field(
        None,
        allow_mutation=False,
        description='The **cases** for the Tag.',
        read_only=True,
        title='cases',
    )
    description: str | None = Field(
        None,
        description='A brief description of the Tag.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='description',
    )
    groups: 'GroupsModel' = Field(
        None,
        allow_mutation=False,
        description='The **groups** for the Tag.',
        read_only=True,
        title='groups',
    )
    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    indicators: 'IndicatorsModel' = Field(
        None,
        allow_mutation=False,
        description='The **indicators** for the Tag.',
        read_only=True,
        title='indicators',
    )
    last_used: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the Tag was last used.',
        read_only=True,
        title='lastUsed',
    )
    name: str | None = Field(
        None,
        description='The **name** for the Tag.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='name',
    )
    normalized: bool = Field(
        None,
        allow_mutation=False,
        description=(
            'Indicates whether this tag is specified as a Main Tag within Tag Normalization.'
        ),
        read_only=True,
        title='normalized',
    )
    owner: str | None = Field(
        None,
        description='The name of the Owner of the Tag.',
        methods=['POST'],
        read_only=False,
        title='owner',
    )
    platforms: dict | None = Field(
        None,
        allow_mutation=False,
        description='For ATT&CK-based tags, these are the platforms applicable to the technique.',
        read_only=True,
        title='platforms',
    )
    security_coverage: dict | None = Field(
        None,
        description=(
            'For ATT&CK-based tags, this is the security coverage level assigned to the tag.'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='securityCoverage',
    )
    synonymous_tag_names: dict | None = Field(
        None,
        allow_mutation=False,
        description=(
            'For Normalized tags, this is a list of defined synonymous tag names that would '
            'normalize to this main tag.'
        ),
        read_only=True,
        title='synonymousTagNames',
    )
    tactics: dict | None = Field(
        None,
        allow_mutation=False,
        description='For ATT&CK-based tags, these are the tactics applicable to the technique.',
        read_only=True,
        title='tactics',
    )
    technique_id: str | None = Field(
        None,
        allow_mutation=False,
        description='For ATT&CK-based tags, this is the technique ID assigned to the tag.',
        read_only=True,
        title='techniqueId',
    )
    victims: 'VictimsModel' = Field(
        None,
        allow_mutation=False,
        description='The **victims** for the Tag.',
        read_only=True,
        title='victims',
    )

    @validator('cases', always=True, pre=True)
    @classmethod
    def _validate_cases(cls, v):
        if not v:
            return CasesModel()  # type: ignore
        return v

    @validator('groups', always=True, pre=True)
    @classmethod
    def _validate_groups(cls, v):
        if not v:
            return GroupsModel()  # type: ignore
        return v

    @validator('indicators', always=True, pre=True)
    @classmethod
    def _validate_indicators(cls, v):
        if not v:
            return IndicatorsModel()  # type: ignore
        return v

    @validator('victims', always=True, pre=True)
    @classmethod
    def _validate_victims(cls, v):
        if not v:
            return VictimsModel()  # type: ignore
        return v


class TagDataModel(
    BaseModel,
    title='Tag Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Tags Data Model"""

    data: list[TagModel] | None = Field(
        [],
        description='The data for the Tags.',
        methods=['POST', 'PUT'],
        title='data',
    )


class TagsModel(
    BaseModel,
    title='Tags Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Tags Model"""

    _mode_support = PrivateAttr(default=True)

    data: list[TagModel] | None = Field(
        [],
        description='The data for the Tags.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.cases.case_model import CasesModel
from tcex.api.tc.v3.groups.group_model import GroupsModel
from tcex.api.tc.v3.indicators.indicator_model import IndicatorsModel
from tcex.api.tc.v3.victims.victim_model import VictimsModel

# add forward references
TagDataModel.update_forward_refs()
TagModel.update_forward_refs()
TagsModel.update_forward_refs()



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Field

# first-party
from tcex.input.field_type.sensitive import Sensitive


class PlaybookCommonModel(BaseModel):
    """Playbook Common Model

    Supported for the following runtimeLevel:
    * ApiService
    * Playbook
    * TriggerService
    * WebhookTriggerService
    """

    tc_cache_kvstore_id: int = Field(
        10,
        description='The KV Store cache DB Id.',
        inclusion_reason='runtimeLevel',
    )
    tc_kvstore_host: str = Field(
        'localhost',
        description='The KV Store hostname.',
        inclusion_reason='runtimeLevel',
    )
    tc_kvstore_pass: Sensitive | None = Field(
        None,
        description='The KV Store password.',
        inclusion_reason='runtimeLevel',
    )
    tc_kvstore_port: int = Field(
        6379,
        description='The KV Store port number.',
        inclusion_reason='runtimeLevel',
    )
    tc_kvstore_tls_enabled: bool = Field(
        default=False,
        description='If true, KV Store requires SSL connection.',
        inclusion_reason='runtimeLevel',
    )
    tc_kvstore_tls_port: int = Field(
        6379,
        description='The KV Store TLS port number.',
        inclusion_reason='runtimeLevel',
    )
    tc_kvstore_type: str = Field(
        'Redis',
        description='The KV Store type (Redis or TCKeyValueAPI).',
        inclusion_reason='runtimeLevel',
    )
    tc_kvstore_user: str | None = Field(
        None,
        description='The KV Store username.',
        inclusion_reason='runtimeLevel',
    )
    tc_playbook_kvstore_id: int = Field(
        0,
        description='The KV Store playbook DB Id.',
        inclusion_reason='runtimeLevel',
    )



"""TcEx Framework Module"""

# standard library
import base64
import json
import logging
import os
from collections.abc import Callable, Iterable
from typing import Any

# third-party
from pydantic import BaseModel

from ...app.key_value_store import KeyValueRedis
from ...app.key_value_store.key_value_store import KeyValueStore
from ...util.util import Util

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


class PlaybookCreate:
    """Playbook Write ABC"""

    def __init__(
        self,
        context: str,
        key_value_store: KeyValueStore,
        output_variables: list,
    ):
        """Initialize the class properties."""
        self.context = context
        self.key_value_store = key_value_store
        self.output_variables = output_variables

        # properties
        self.log = _logger
        self.util = Util()

    @staticmethod
    def _check_iterable(value: dict | Iterable | str, validate: bool):
        """Raise an exception if value is not an Iterable.

        Validation:
          - not a dict (dicts are iterable)
          - not a string (strings are iterable)
          - is Iterable
        """
        if validate is True and (isinstance(value, dict | str) or not isinstance(value, Iterable)):
            ex_msg = 'Invalid data provided for KeyValueArray.'
            raise RuntimeError(ex_msg)

    def _check_null(self, key: str, value: Any) -> bool:
        """Return True if key or value is null."""
        invalid = False

        # this code should be unreachable, but just in case
        if key is None:
            self.log.warning('The provided key was None.')
            invalid = True

        if value is None:
            self.log.warning(f'The provided value for key {key} was None.')
            invalid = True

            # specifically to allow the tcex-test framework to validate outputs
            if os.getenv('TC_PLAYBOOK_WRITE_NULL') is not None and isinstance(
                self.key_value_store.client, KeyValueRedis
            ):
                variable = self._get_variable(key)
                self.log.debug(f'event=writing-null-to-kvstore, variable={variable}')
                self.key_value_store.redis_client.hset(
                    self.context, f'{variable}_NULL_VALIDATION', ''
                )

        return invalid

    def _check_requested(self, variable: str, when_requested: bool) -> bool:
        """Return True if output variable was requested by downstream app."""
        if when_requested is True and not self.is_requested(variable):
            self.log.debug(f'Variable {variable} was NOT requested by downstream app.')
            return False
        return True

    def _check_variable_type(self, variable: str, type_: str):
        """Validate the correct type was passed to the method."""
        if self.util.get_playbook_variable_type(variable).lower() != type_.lower():
            ex_msg = f'Invalid variable provided ({variable}), variable must be of type {type_}.'
            raise RuntimeError(ex_msg)

    @staticmethod
    def _coerce_string_value(value: bool | float | str) -> str:
        """Return a string value from an bool or int."""
        # coerce bool before int as python says a bool is an int
        if isinstance(value, bool):
            # coerce bool to str type
            value = str(value).lower()

        # coerce int to str type
        if isinstance(value, float | int):
            value = str(value)

        return value

    def _create_data(self, key: str, value: bytes | str) -> int | None:
        """Write data to key value store."""
        self.log.debug(f'writing variable {key.strip()}')
        try:
            return self.key_value_store.client.create(self.context, key.strip(), value)
        except RuntimeError:  # pragma: no cover
            self.log.exception('Error writing data to key value store.')
            return None

    def _get_variable(self, key: str, variable_type: str | None = None) -> str | None:
        """Return properly formatted variable.

        A key can be provided as the variable key (e.g., app.output) or the
        entire (e.g., #App:1234:app.output!String). The full variable is required
        to create the record in the KV Store.

        If a variable_type is provided an exact match will be found, however if no
        variable type is known the first key to match will be returned. Uniqueness
        of keys is not guaranteed, but in more recent Apps it is the standard.

        If no variable is found it means that the variable was not requested by the
        any downstream Apps or could possible be formatted incorrectly.
        """
        if not self.util.is_playbook_variable(key):
            # try to lookup the variable in the requested output variables.
            for output_variable in self.output_variables:
                variable_model = self.util.get_playbook_variable_model(output_variable)
                if (
                    variable_model
                    and variable_model.key == key
                    and (variable_type is None or variable_model.type == variable_type)
                ):
                    # either an exact match, or first match
                    return output_variable

            # not requested by downstream App or misconfigured
            return None

        # key was already a properly formatted variable
        return key

    @staticmethod
    def _serialize_data(value: dict | list | str) -> str:
        """Get the value from Redis if applicable."""
        try:
            return json.dumps(value)
        except ValueError as ex:  # pragma: no cover
            ex_msg = f'Invalid data provided, failed to serialize value ({ex}).'
            raise RuntimeError(ex_msg) from ex

    @staticmethod
    def _process_object_types(
        value: BaseModel | dict,
        validate: bool = True,
        allow_none: bool = False,
    ) -> dict[str, Any]:
        """Process object types (e.g., KeyValue, TCEntity)."""
        types = (BaseModel, dict)
        if allow_none is True:
            types = (BaseModel, dict, type(None))

        if validate and not isinstance(value, types):
            ex_msg = f'Invalid type provided for object type ({type(value)}).'
            raise RuntimeError(ex_msg)

        if isinstance(value, BaseModel):
            value = value.dict(exclude_unset=True)

        return value

    @staticmethod
    def is_key_value(data: dict) -> bool:
        """Return True if provided data has proper structure for Key Value."""
        if not isinstance(data, dict):
            return False
        return all(x in data for x in ['key', 'value'])

    def is_requested(self, variable: str) -> bool:
        """Return True if provided variable was requested by downstream App."""
        return variable in self.output_variables

    @staticmethod
    def is_tc_batch(data: dict) -> bool:
        """Return True if provided data has proper structure for TC Batch."""
        if not isinstance(data, dict):
            return False
        if not isinstance(data.get('indicator', []), list):
            return False
        return isinstance(data.get('group', []), list)

    @staticmethod
    def is_tc_entity(data: dict) -> bool:
        """Return True if provided data has proper structure for TC Entity."""
        if not isinstance(data, dict):
            return False
        return all(x in data for x in ['id', 'value', 'type'])

    def any(
        self,
        key: str,
        value: (
            BaseModel | bytes | dict | str | list[BaseModel] | list[bytes] | list[dict] | list[str]
        ),
        validate: bool = True,
        variable_type: str | None = None,
        when_requested: bool = True,
    ) -> int | None:
        """Write the value to the keystore for all types.

        This is a quick helper method, for more advanced features
        the individual write methods should be used (e.g., binary).

        Args:
            key: The variable to write to the DB (e.g., app.colors).
            value: The data to write to the DB.
            validate: Perform validation on the data.
            variable_type: The variable type being written. Only required if not unique.
            when_requested: Only write the data if the variable was requested by downstream App.
        """
        if self._check_null(key, value) is True:
            return None

        # convert key to variable if required
        variable = self._get_variable(key, variable_type)
        if variable is None:
            # variable is invalid or not requested by downstream App
            return None

        if self._check_requested(variable, when_requested) is False:
            # variable is not requested by downstream App
            return None

        # get the type from the variable
        variable_type = self.util.get_playbook_variable_type(variable).lower()

        # map type to create method
        variable_type_map: dict[str, Callable] = {
            'binary': self.binary,
            'binaryarray': self.binary_array,
            'keyvalue': self.key_value,
            'keyvaluearray': self.key_value_array,
            'string': self.string,
            'stringarray': self.string_array,
            'tcentity': self.tc_entity,
            'tcentityarray': self.tc_entity_array,
            'tcbatch': self.tc_batch,
        }
        return variable_type_map.get(variable_type, self.raw)(  # type: ignore
            variable,
            value,
            validate,
            when_requested,  # type: ignore
        )

    def binary(
        self,
        key: str,
        value: bytes,
        validate: bool = True,
        when_requested: bool = True,
    ) -> int | None:
        """Create the value in Redis if applicable."""
        if self._check_null(key, value) is True:
            return None

        # convert key to variable if required
        variable = self._get_variable(key, 'Binary')
        if variable is None:
            # variable is invalid or not requested by downstream App
            return None

        if self._check_requested(variable, when_requested) is False:
            # variable is not requested by downstream App
            return None

        # quick check to ensure an invalid type was not provided
        self._check_variable_type(variable, 'Binary')

        # basic validation of value
        if validate and not isinstance(value, bytes):
            ex_msg = 'Invalid data provided for Binary.'
            raise RuntimeError(ex_msg)

        # prepare value - playbook Binary fields are base64 encoded
        value_ = base64.b64encode(value).decode('utf-8')
        value_ = self._serialize_data(value_)
        return self._create_data(variable, value_)

    def binary_array(
        self,
        key: str,
        value: list[bytes],
        validate: bool = True,
        when_requested: bool = True,
    ) -> int | None:
        """Create the value in Redis if applicable."""
        if self._check_null(key, value) is True:
            return None

        # validate array type provided
        self._check_iterable(value, validate)

        # convert key to variable if required
        variable = self._get_variable(key, 'BinaryArray')
        if variable is None:
            return None

        if self._check_requested(variable, when_requested) is False:
            return None

        # quick check to ensure an invalid type was not provided
        self._check_variable_type(variable, 'BinaryArray')

        # basic validation and prep of value
        value_encoded = []
        for v in value:
            v_ = v
            if v_ is not None:
                if validate and not isinstance(v_, bytes):
                    ex_msg = 'Invalid data provided for Binary.'
                    raise RuntimeError(ex_msg)
                v_ = base64.b64encode(v_).decode('utf-8')
            value_encoded.append(v_)
        return self._create_data(variable, self._serialize_data(value_encoded))

    def key_value(
        self,
        key: str,
        value: BaseModel | dict,
        validate: bool = True,
        when_requested: bool = True,
    ) -> int | None:
        """Create the value in Redis if applicable."""
        if self._check_null(key, value) is True:
            return None

        # convert key to variable if required
        variable = self._get_variable(key, 'KeyValue')
        if variable is None:
            return None

        if self._check_requested(variable, when_requested) is False:
            return None

        # quick check to ensure an invalid type was not provided
        self._check_variable_type(variable, 'KeyValue')

        # basic validation and prep of value
        value = self._process_object_types(value, validate)
        if validate and not self.is_key_value(value):
            ex_msg = 'Invalid data provided for KeyValueArray.'
            raise RuntimeError(ex_msg)

        return self._create_data(variable, self._serialize_data(value))

    def key_value_array(
        self,
        key: str,
        value: list[BaseModel | dict],
        validate: bool = True,
        when_requested: bool = True,
    ):
        """Create the value in Redis if applicable."""
        if self._check_null(key, value) is True:
            return None

        # validate array type provided
        self._check_iterable(value, validate)

        # convert key to variable if required
        variable = self._get_variable(key, 'KeyValueArray')
        if variable is None:
            return None

        if self._check_requested(variable, when_requested) is False:
            return None

        # quick check to ensure an invalid type was not provided
        self._check_variable_type(variable, 'KeyValueArray')

        # basic validation and prep of value
        _value = []
        for v in value:
            v_ = self._process_object_types(v, validate, allow_none=True)
            if validate and not self.is_key_value(v_):
                ex_msg = 'Invalid data provided for KeyValueArray.'
                raise RuntimeError(ex_msg)
            _value.append(v_)
        value = _value

        return self._create_data(variable, self._serialize_data(value))

    def string(
        self,
        key: str,
        value: bool | float | str,
        validate: bool = True,
        when_requested: bool = True,
    ) -> int | None:
        """Create the value in Redis if applicable."""
        if self._check_null(key, value) is True:
            return None

        # convert key to variable if required
        variable = self._get_variable(key, 'String')
        if variable is None:
            return None

        if self._check_requested(variable, when_requested) is False:
            return None

        # quick check to ensure an invalid type was not provided
        self._check_variable_type(variable, 'String')

        # coerce string values
        value = self._coerce_string_value(value)

        # validation only needs to check str because value was coerced
        if validate and not isinstance(value, str):
            ex_msg = f'Invalid data provided for String ({value}).'
            raise RuntimeError(ex_msg)

        return self._create_data(variable, self._serialize_data(value))

    def string_array(
        self,
        key: str,
        value: list[bool | float | int | str],
        validate: bool = True,
        when_requested: bool = True,
    ):
        """Create the value in Redis if applicable."""
        if self._check_null(key, value) is True:
            return None

        # validate array type provided
        self._check_iterable(value, validate)

        # convert key to variable if required
        variable = self._get_variable(key, 'StringArray')
        if variable is None:
            return None

        if self._check_requested(variable, when_requested) is False:
            return None

        # quick check to ensure an invalid type was not provided
        self._check_variable_type(variable, 'StringArray')

        # basic validation and prep of value
        value_coerced = []
        for v in value:
            # coerce string values
            v_ = self._coerce_string_value(v)

            # validation only needs to check str because value was coerced
            if validate and not isinstance(v_, type(None) | str):
                ex_msg = 'Invalid data provided for StringArray.'
                raise RuntimeError(ex_msg)
            value_coerced.append(v_)
        value = value_coerced

        return self._create_data(variable, self._serialize_data(value))

    def raw(
        self,
        key: str,
        value: bytes | str,
        _validate: bool = True,
        _when_requested: bool = True,
    ) -> int | None:
        """Create method of CRUD operation for raw data.

        Raw data can only be a byte, str or int. Other data
        structures (dict, list, etc) must be serialized.
        """
        if self._check_null(key, value):
            return None

        return self._create_data(key, value)

    def tc_batch(
        self,
        key: str,
        value: BaseModel | dict,
        validate: bool = True,
        when_requested: bool = True,
    ) -> int | None:
        """Create the value in Redis if applicable."""
        if self._check_null(key, value) is True:
            return None

        # convert key to variable if required
        variable = self._get_variable(key, 'TCBatch')
        if variable is None:
            return None

        if self._check_requested(variable, when_requested) is False:
            return None

        # quick check to ensure an invalid type was not provided
        self._check_variable_type(variable, 'TCBatch')

        # basic validation
        value = self._process_object_types(value, validate)
        if validate and not self.is_tc_batch(value):
            ex_msg = 'Invalid data provided for TcBatch.'
            raise RuntimeError(ex_msg)

        return self._create_data(variable, self._serialize_data(value))

    def tc_entity(
        self,
        key: str,
        value: BaseModel | dict,
        validate: bool = True,
        when_requested: bool = True,
    ) -> int | None:
        """Create the value in Redis if applicable."""
        if self._check_null(key, value) is True:
            return None

        # convert key to variable if required
        variable = self._get_variable(key, 'TCEntity')
        if variable is None:
            return None

        if self._check_requested(variable, when_requested) is False:
            return None

        # quick check to ensure an invalid type was not provided
        self._check_variable_type(variable, 'TCEntity')

        # basic validation
        value = self._process_object_types(value, validate)
        if validate and not self.is_tc_entity(value):
            ex_msg = 'Invalid data provided for TcEntityArray.'
            raise RuntimeError(ex_msg)

        return self._create_data(variable, self._serialize_data(value))

    def tc_entity_array(
        self,
        key: str,
        value: list[BaseModel | dict],
        validate: bool = True,
        when_requested: bool = True,
    ):
        """Create the value in Redis if applicable."""
        if self._check_null(key, value) is True:
            return None

        # validate array type provided
        self._check_iterable(value, validate)

        # convert key to variable if required
        variable = self._get_variable(key, 'TCEntityArray')
        if variable is None:
            return None

        if self._check_requested(variable, when_requested) is False:
            return None

        # quick check to ensure an invalid type was not provided
        self._check_variable_type(variable, 'TCEntityArray')

        # basic validation and prep of value
        _value = []
        for v in value:
            v_ = self._process_object_types(v, validate, allow_none=True)
            if validate and not self.is_tc_entity(v_):
                ex_msg = 'Invalid data provided for TcEntityArray.'
                raise RuntimeError(ex_msg)
            _value.append(v_)
        value = _value

        return self._create_data(variable, self._serialize_data(value))

    def variable(
        self,
        key: str,
        value: (
            BaseModel
            | bytes
            | dict
            | int
            | str
            | list
            | list[BaseModel | None]
            | list[bytes | None]
            | list[dict | None]
            | list[str | None]
            | None
        ),
        variable_type: str | None = None,
    ) -> int | None:
        """Alias for any method of CRUD operation for working with KeyValue DB.

        This method will automatically check to see if provided variable was requested by
        a downstream app and if so create the data in the KeyValue DB.

        Args:
            key: The variable to write to the DB (e.g., app.colors).
            value: The data to write to the DB.
            variable_type: The variable type being written. Only required if not unique.
        """
        if self._check_null(key, value) is True:
            return None

        # short-circuit the process, if there are no downstream variables requested.
        if not self.output_variables:  # pragma: no cover
            self.log.debug(f'Variable {key} was NOT requested by downstream app.')
            return None

        # key can be provided as the variable key (e.g., app.output) or
        # the entire (e.g., #App:1234:app.output!String). we need the
        # full variable to proceed.
        variable = self._get_variable(key, variable_type)
        if variable is None:
            return None

        if variable is None or variable not in self.output_variables:
            self.log.debug(f'Variable {key} was NOT requested by downstream app.')
            return None

        # write the variable (None value would be caught in _check_null method)
        return self.any(variable, value)  # type: ignore



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Field

# first-party
from tcex.input.field_type.sensitive import Sensitive


class CalSettingModel(BaseModel):
    """CAL Settings Model

    Feature: CALSettings

    Supported for the following runtimeLevel:
    * ApiService
    * Playbook
    * WebhookTriggerService
    * TriggerService
    """

    #
    # ThreatConnect Provided Inputs
    #

    tc_cal_host: str = Field(
        ...,
        description='The hostname for CAL.',
        inclusion_reason='feature (CALSettings)',
    )
    tc_cal_token: Sensitive = Field(
        ...,
        description='The token for CAL.',
        inclusion_reason='feature (CALSettings)',
    )
    tc_cal_timestamp: int = Field(
        ...,
        description='The expiration timestamp in epoch for tc_cal_token.',
        inclusion_reason='feature (CALSettings)',
    )



"""TcEx Framework Module"""

# standard library
import threading
from typing import ClassVar


class Singleton(type):
    """A singleton Metaclass"""

    _instances: ClassVar = {}
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        """Evoke call method."""
        with cls._lock:
            if cls not in cls._instances:
                cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]



"""TcEx Framework Module"""

# standard library
import logging

# third-party
from requests import Session  # TYPE-CHECKING

# first-party
from tcex.exit.error_code import handle_error
from tcex.logger.trace_logger import TraceLogger
from tcex.util import Util

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class Metric:
    """TcEx Metrics Class

    Args:
        session_tc: An configured instance of request.Session with TC API Auth.
        name: The name for the metric.
        description: The description of the metric.
        data_type: The type of metric: Sum, Count, Min, Max, First, Last, and Average.
        interval: The metric interval: Hourly, Daily, Weekly, Monthly, and Yearly.
        keyed: Indicates whether the data will have a keyed value.
    """

    def __init__(
        self,
        session_tc: Session,
        name: str,
        description: str,
        data_type: str,
        interval: str,
        keyed: bool = False,
    ):
        """Initialize the Class properties."""
        self.session_tc = session_tc
        self._metric_data_type = data_type
        self._metric_description = description
        self._metric_interval = interval
        self._metric_keyed = keyed
        self._metric_name = name

        # properties
        self._metric_id = None
        self.log = _logger
        self.util = Util

        if not self.metric_find():
            self.metric_create()

    def metric_create(self):
        """Create the defined metric.

        .. code-block:: javascript

            {
                "status": "Success",
                "data": {
                    "customMetricConfig": {
                        "id": 12,
                        "name": "Added Reports",
                        "dataType": "Sum",
                        "interval": "Daily",
                        "keyedValues": false,
                        "description": "Added reports daily count."
                    }
                }
            }
        """
        body = {
            'dataType': self._metric_data_type,
            'description': self._metric_description,
            'interval': self._metric_interval,
            'name': self._metric_name,
            'keyedValues': self._metric_keyed,
        }
        self.log.debug(f'metric body: {body}')
        r = self.session_tc.post('/v2/customMetrics', json=body)

        if not r.ok:  # pragma: no cover
            handle_error(700, [r.status_code, r.text])

        data = r.json()
        self._metric_id = data.get('data', {}).get('customMetricConfig', {}).get('id')
        self.log.debug(f'metric data: {data}')

    def metric_find(self):
        """Find the Metric by name.

        .. code-block:: javascript

            {
                "status": "Success",
                "data": {
                    "resultCount": 1,
                    "customMetricConfig": [
                        {
                            "id": 9,
                            "name": "My Metric",
                            "dataType": "Sum",
                            "interval": "Daily",
                            "keyedValues": false,
                            "description": "TcEx Metric Testing"
                        }
                    ]
                }
            }
        """
        params: dict[str, int] = {'resultLimit': 50, 'resultStart': 0}
        while True:
            if params['resultStart'] >= params['resultLimit']:
                break
            r = self.session_tc.get('/v2/customMetrics', params=params)
            if not r.ok:  # pragma: no cover
                handle_error(705, [r.status_code, r.text])
            data = r.json()
            for metric in data.get('data', {}).get('customMetricConfig'):
                if metric.get('name') == self._metric_name:
                    self._metric_id = metric.get('id')
                    self.log.info(
                        f'found metric with name "{self._metric_name}" and Id {self._metric_id}.'
                    )
                    return True
            params['resultStart'] += params['resultLimit']
        return False

    def add(self, value, date=None, return_value=False, key=None, weight=None):
        """Add metric data to collection.

        Args:
            value (str): The value of the metric.
            date (str, optional): The optional date of the metric.
            return_value (bool, default:False): Tell the API to return the updates metric value.
            key (str, optional): The key value for keyed metric.
            weight (str, optional): The weight value (only needed for averages)

        Return:
            dict: If return_value is True a dict with the current value for the time period
                is returned.
        """
        data = {}
        http_success_code = 200
        http_no_content_code = 204
        if self._metric_id is None:  # pragma: no cover
            handle_error(715, [self._metric_name])

        body = {'value': value}
        if date is not None:
            body['date'] = self.util.any_to_datetime(date).strftime('%Y-%m-%dT%H:%M:%SZ')

        if key is not None:
            body['name'] = key
        if weight:
            body['weight'] = weight
        self.log.debug(f'metric data: {body}')

        params = {}
        if return_value:
            params = {'returnValue': 'true'}

        url = f'/v2/customMetrics/{self._metric_id}/data'
        r = self.session_tc.post(url, json=body, params=params)
        if r.status_code == http_success_code and 'application/json' in r.headers.get(
            'content-type', ''
        ):
            data = r.json()
        elif r.status_code == http_no_content_code:
            pass
        else:  # pragma: no cover
            handle_error(710, [r.status_code, r.text])

        return data

    def add_keyed(self, value, key, date=None, return_value=False, weight=None):
        """Add keyed metric data to collection.

        Args:
            value (str): The value of the metric.
            key (str): The key value for keyed metric.
            date (str, optional): The optional date of the metric.
            return_value (bool, default:False): Tell the API to return the updates metric value.
            weight (str, optional): The weight value (only needed for averages)

        Return:
            dict: If return_value is True a dict with the current value for the time period
                is returned.
        """
        return self.add(value, date, return_value, key, weight)



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Extra, validator
from pydantic.fields import ModelField  # TYPE-CHECKING

# first-party
from tcex.input.field_type.exception import InvalidEmptyValue


class TCEntity(BaseModel):
    """Model for TCEntity Input."""

    id: int
    type: str
    value: str

    # IMPORTANT: confidence and rating values are included only so that, when defined,
    #    they come back as the correct types. They are only intended for Indicator
    #    types and are not guaranteed to be populated.
    confidence: int | None
    rating: int | None

    @validator('id', 'type', 'value')
    @classmethod
    def non_empty_string(cls, v: dict[str, str], field: ModelField) -> dict[str, str]:
        """Validate that the value is a non-empty string."""
        if isinstance(v, str) and v.replace(' ', '') == '':  # None value are automatically covered
            raise InvalidEmptyValue(field_name=field.name)
        return v

    class Config:
        """Model Config"""

        extra = Extra.allow
        validate_assignment = True



"""TcEx Framework Module"""

# standard library
from typing import Any
from urllib.parse import quote

# third-party
from requests import Session

from .key_value_abc import KeyValueABC


class KeyValueApi(KeyValueABC):
    """TcEx Key Value API Module.

    Args:
        session: A configured requests session for TC API (tcex.session).
    """

    def __init__(self, session: Session):
        """Initialize the Class properties."""
        self._session = session

        # properties
        self.kv_type = 'api'

    def create(self, context: str, key: str, value: Any) -> int:
        """Create key/value pair in remote KV store.

        Args:
            context: A specific context for the create.
            key: The key to create in remote KV store.
            value: The value to store in remote KV store.

        Returns:
            (string): The response from the API call.
        """
        key = quote(key, safe='~')
        headers = {'content-type': 'application/octet-stream'}
        url = f'/internal/playbooks/keyValue/{context}/{key}'

        r = self._session.put(url, data=value, headers=headers)
        if r.ok:
            # the redis client returns the count of items added, in this case there is only one
            return 1
        return 0  # no entries created

    def read(self, context: str, key: str) -> bytes | str | None:
        """Read data from remote KV store for the provided key.

        Args:
            context: A specific context for the create.
            key: The key to read in remote KV store.

        Returns:
            (any): The response data from the remote KV store.
        """
        key = quote(key, safe='~')

        url = f'/internal/playbooks/keyValue/{context}/{key}'
        r = self._session.get(url)
        data = r.content

        # Binary data for PB Apps is base64 encoded, for service Apps it is not
        if data is not None and isinstance(data, bytes):
            data = data.decode('utf-8')
        return data



"""TcEx Framework Module"""

# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class NoteModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='Note Model',
    validate_assignment=True,
):
    """Note Model"""

    _associated_type = PrivateAttr(default=False)
    _cm_type = PrivateAttr(default=True)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    artifact: 'ArtifactModel' = Field(
        None,
        allow_mutation=False,
        description='The **artifact** for the Note.',
        read_only=True,
        title='artifact',
    )
    artifact_id: int | None = Field(
        None,
        description='The ID of the Artifact on which to apply the Note.',
        methods=['POST'],
        read_only=False,
        title='artifactId',
    )
    author: str | None = Field(
        None,
        allow_mutation=False,
        description='The **author** for the Note.',
        read_only=True,
        title='author',
    )
    case_id: int | None = Field(
        None,
        description='The **case id** for the Note.',
        methods=['POST'],
        read_only=False,
        required_alt_field='caseXid',
        title='caseId',
    )
    case_xid: str | None = Field(
        None,
        description='The **case xid** for the Note.',
        methods=['POST'],
        read_only=False,
        required_alt_field='caseId',
        title='caseXid',
    )
    date_added: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The **date added** for the Note.',
        read_only=True,
        title='dateAdded',
    )
    edited: bool = Field(
        None,
        allow_mutation=False,
        description='The **edited** for the Note.',
        read_only=True,
        title='edited',
    )
    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    last_modified: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The **last modified** for the Note.',
        read_only=True,
        title='lastModified',
    )
    parent_case: 'CaseModel' = Field(
        None,
        allow_mutation=False,
        description='The **parent case** for the Note.',
        read_only=True,
        title='parentCase',
    )
    summary: str | None = Field(
        None,
        allow_mutation=False,
        description='The **summary** for the Note.',
        read_only=True,
        title='summary',
    )
    task: 'TaskModel' = Field(
        None,
        allow_mutation=False,
        description='The **task** for the Note.',
        read_only=True,
        title='task',
    )
    task_id: int | None = Field(
        None,
        description='The ID of the Task on which to apply the Note.',
        methods=['POST'],
        read_only=False,
        title='taskId',
    )
    task_xid: str | None = Field(
        None,
        description='The XID of the Task on which to apply the Note.',
        methods=['POST'],
        read_only=False,
        title='taskXid',
    )
    text: str | None = Field(
        None,
        description='The **text** for the Note.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='text',
    )
    workflow_event: 'WorkflowEventModel' = Field(
        None,
        allow_mutation=False,
        description='The **workflow event** for the Note.',
        read_only=True,
        title='workflowEvent',
    )
    workflow_event_id: int | None = Field(
        None,
        description='The ID of the Event on which to apply the Note.',
        methods=['POST'],
        read_only=False,
        title='workflowEventId',
    )

    @validator('artifact', always=True, pre=True)
    @classmethod
    def _validate_artifact(cls, v):
        if not v:
            return ArtifactModel()  # type: ignore
        return v

    @validator('parent_case', always=True, pre=True)
    @classmethod
    def _validate_case(cls, v):
        if not v:
            return CaseModel()  # type: ignore
        return v

    @validator('task', always=True, pre=True)
    @classmethod
    def _validate_task(cls, v):
        if not v:
            return TaskModel()  # type: ignore
        return v

    @validator('workflow_event', always=True, pre=True)
    @classmethod
    def _validate_workflow_event(cls, v):
        if not v:
            return WorkflowEventModel()  # type: ignore
        return v


class NoteDataModel(
    BaseModel,
    title='Note Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Notes Data Model"""

    data: list[NoteModel] | None = Field(
        [],
        description='The data for the Notes.',
        methods=['POST', 'PUT'],
        title='data',
    )


class NotesModel(
    BaseModel,
    title='Notes Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Notes Model"""

    _mode_support = PrivateAttr(default=False)

    data: list[NoteModel] | None = Field(
        [],
        description='The data for the Notes.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactModel
from tcex.api.tc.v3.cases.case_model import CaseModel
from tcex.api.tc.v3.tasks.task_model import TaskModel
from tcex.api.tc.v3.workflow_events.workflow_event_model import WorkflowEventModel

# add forward references
NoteDataModel.update_forward_refs()
NoteModel.update_forward_refs()
NotesModel.update_forward_refs()



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class ExclusionListModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='ExclusionList Model',
    validate_assignment=True,
):
    """Exclusion_List Model"""

    _associated_type = PrivateAttr(default=False)
    _cm_type = PrivateAttr(default=False)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    active: bool = Field(
        None,
        description='Whether the rule is active or not.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='active',
    )
    description: str | None = Field(
        None,
        allow_mutation=False,
        description='The **description** for the Exclusion_List.',
        read_only=True,
        title='description',
    )
    fixed_values: dict | None = Field(
        None,
        description='A list of exclusions represent fixed values.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='fixedValues',
    )
    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    managed: bool = Field(
        None,
        description='Whether the rule is managed by the system.',
        methods=['POST'],
        read_only=False,
        title='managed',
    )
    name: str | None = Field(
        None,
        description='The name of the rule.',
        methods=['POST'],
        read_only=False,
        title='name',
    )
    owner: str | None = Field(
        None,
        description='The name of the Owner of the Exclusion List.',
        methods=['POST'],
        read_only=False,
        title='owner',
    )
    owner_id: int | None = Field(
        None,
        description='The ID of the owner of this exclusion list.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='ownerId',
    )
    variable_values: dict | None = Field(
        None,
        description='A list of exclusions serve as wildcards.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='variableValues',
    )


class ExclusionListDataModel(
    BaseModel,
    title='ExclusionList Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Exclusion_Lists Data Model"""

    data: list[ExclusionListModel] | None = Field(
        [],
        description='The data for the ExclusionLists.',
        methods=['POST', 'PUT'],
        title='data',
    )


class ExclusionListsModel(
    BaseModel,
    title='ExclusionLists Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Exclusion_Lists Model"""

    _mode_support = PrivateAttr(default=False)

    data: list[ExclusionListModel] | None = Field(
        [],
        description='The data for the ExclusionLists.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# add forward references
ExclusionListDataModel.update_forward_refs()
ExclusionListModel.update_forward_refs()
ExclusionListsModel.update_forward_refs()



"""TcEx Framework Module"""

# standard library
import json
import uuid
from collections.abc import Callable
from typing import ForwardRef

# first-party
from tcex.api.tc.v2.batch.attribute import Attribute
from tcex.api.tc.v2.batch.security_label import SecurityLabel
from tcex.api.tc.v2.batch.tag import Tag
from tcex.util import Util

FileOccurrences = ForwardRef('FileOccurrences')

# import local modules for dynamic reference
module = __import__(__name__)


def custom_indicator_class_factory(indicator_type, base_class, class_dict, value_fields):
    """Return internal methods for dynamically building Custom Indicator Class."""
    value_count = len(value_fields)

    def init_1(self, tcex, value1, xid, **kwargs):
        """Init method for Custom Indicator Types with one value"""
        summary = self.build_summary(value1)  # build the indicator summary
        base_class.__init__(self, tcex, indicator_type, summary, xid, **kwargs)
        for k, v in class_dict.items():
            setattr(self, k, v)

    def init_2(self, tcex, value1, value2, xid, **kwargs):
        """Init method for Custom Indicator Types with two values."""
        summary = self.build_summary(value1, value2)  # build the indicator summary
        base_class.__init__(self, tcex, indicator_type, summary, xid, **kwargs)
        for k, v in class_dict.items():
            setattr(self, k, v)

    def init_3(self, tcex, value1, value2, value3, xid, **kwargs):
        """Init method for Custom Indicator Types with three values."""
        summary = self.build_summary(value1, value2, value3)  # build the indicator summary
        base_class.__init__(self, tcex, indicator_type, summary, xid, **kwargs)
        for k, v in class_dict.items():
            setattr(self, k, v)

    class_name = indicator_type.replace(' ', '')
    init_method = locals()[f'init_{value_count}']
    return type(str(class_name), (base_class,), {'__init__': init_method})


class Indicator:
    """ThreatConnect Batch Indicator Object"""

    __slots__ = [
        '_attributes',
        '_file_actions',
        '_indicator_data',
        '_labels',
        '_occurrences',
        '_summary',
        '_tags',
        '_type',
        'util',
    ]

    def __init__(self, indicator_type: str, summary: str, **kwargs):
        """Initialize instance properties.

        Args:
            indicator_type: The ThreatConnect define Indicator type.
            summary: The value for this Indicator.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        self._summary = summary
        self._type = indicator_type
        self._indicator_data: dict[str, bool | dict | float | int | list | str] = {
            'summary': summary,
            'type': indicator_type,
        }

        # properties
        self._attributes = []
        self._file_actions = []
        self._labels = []
        self._occurrences = []
        self._tags = []
        self.util = Util()

        # process all kwargs and update metadata field names
        for arg, value in kwargs.items():
            self.add_key_value(arg, value)
        # set xid to random and unique uuid4 value if not provided
        if kwargs.get('xid') is None:
            self._indicator_data['xid'] = str(uuid.uuid4())

    @property
    def _metadata_map(self) -> dict:
        """Return metadata map for Indicator objects."""
        return {
            'date_added': 'dateAdded',
            'dnsActive': 'flag1',
            'dns_active': 'flag1',
            'last_modified': 'lastModified',
            'private_flag': 'privateFlag',
            'size': 'intValue1',
            'whoisActive': 'flag2',
            'whois_active': 'flag2',
        }

    def add_key_value(self, key: str, value: str):
        """Add custom field to Indicator object.

        .. note:: The key must be the exact name required by the batch schema.

        Example::

            file_hash = tcex.batch.file(
                'File', '1d878cdc391461e392678ba3fc9f6f32'
            )
            file_hash.add_key_value('size', '1024')

        Args:
            key: The field key to add to the JSON batch data.
            value: The field value to add to the JSON batch data.
        """
        key = self._metadata_map.get(key, key)

        if key in {
            'dateAdded',
            'lastModified',
            'firstSeen',
            'lastSeen',
            'externalDateCreated',
            'externalDateExpires',
            'externalLastModified',
        }:
            self._indicator_data[key] = self.util.any_to_datetime(value).strftime(
                '%Y-%m-%dT%H:%M:%SZ'
            )
        elif key == 'confidence':
            self._indicator_data[key] = int(value)
        elif key == 'rating':
            self._indicator_data[key] = float(value)
        else:
            self._indicator_data[key] = value

    @property
    def active(self) -> bool:
        """Return Indicator active."""
        return self._indicator_data.get('active')  # type: ignore

    @active.setter
    def active(self, active: bool):
        """Set Indicator active."""
        self._indicator_data['active'] = self.util.to_bool(active)

    def association(self, group_xid: str):
        """Add association using xid value.

        Args:
            group_xid (str): The external id of the Group to associate.
        """
        association = {'groupXid': group_xid}
        self._indicator_data.setdefault('associatedGroups', []).append(association)  # type: ignore

    def attribute(
        self,
        attr_type: str,
        attr_value: str,
        displayed: bool = False,
        source: str | None = None,
        unique: bool = True,
        formatter: Callable[[str], str] | None = None,
    ) -> Attribute:
        """Return instance of Attribute

        unique:
            * False - Attribute type:value can be duplicated.
            * Type - Attribute type has to be unique (e.g., only 1 Description Attribute).
            * True - Attribute type:value combo must be unique.

        Args:
            attr_type: The ThreatConnect defined attribute type.
            attr_value: The value for this attribute.
            displayed: If True the supported attribute will be marked for display.
            source: The source value for this attribute.
            unique: Control attribute creation.
            formatter: A callable that takes a single attribute
                value and returns a single formatted value.

        Returns:
            Attribute: An instance of the Attribute class.
        """
        attr = Attribute(attr_type, attr_value, displayed, source, formatter)
        if unique == 'Type':
            for attribute_data in self._attributes:
                if attribute_data.type == attr_type:
                    attr = attribute_data
                    break
            else:
                self._attributes.append(attr)
        elif unique is True:
            for attribute_data in self._attributes:
                if attribute_data.type == attr_type and attribute_data.value == attr.value:
                    attr = attribute_data
                    break
            else:
                self._attributes.append(attr)
        elif unique is False:
            self._attributes.append(attr)
        return attr

    @staticmethod
    def build_summary(
        val1: str | None = None, val2: str | None = None, val3: str | None = None
    ) -> str:
        """Build the Indicator summary using available values."""
        summary = []
        if val1 is not None:
            summary.append(val1)
        if val2 is not None:
            summary.append(val2)
        if val3 is not None:
            summary.append(val3)
        return ' : '.join(summary)

    @property
    def confidence(self) -> int:
        """Return Indicator confidence."""
        return self._indicator_data.get('confidence')  # type: ignore

    @confidence.setter
    def confidence(self, confidence: int):
        """Set Indicator confidence."""
        self._indicator_data['confidence'] = int(confidence)

    @property
    def data(self) -> dict:
        """Return Indicator data."""
        # add attributes
        if self._attributes:
            self._indicator_data['attribute'] = []
            for attr in self._attributes:
                if attr.valid:
                    self._indicator_data['attribute'].append(attr.data)
        # add file actions
        if self._file_actions:
            self._indicator_data.setdefault('fileAction', {})
            self._indicator_data['fileAction'].setdefault('children', [])  # type: ignore
            for action in self._file_actions:
                self._indicator_data['fileAction']['children'].append(action.data)  # type: ignore
        # add file occurrences
        if self._occurrences:
            self._indicator_data.setdefault('fileOccurrence', [])
            for occurrence in self._occurrences:
                self._indicator_data['fileOccurrence'].append(occurrence.data)  # type: ignore
        # add security labels
        if self._labels:
            self._indicator_data['securityLabel'] = []
            for label in self._labels:
                self._indicator_data['securityLabel'].append(label.data)
        # add tags
        if self._tags:
            self._indicator_data['tag'] = []
            for tag in self._tags:
                if tag.valid:
                    self._indicator_data['tag'].append(tag.data)
        return self._indicator_data

    @property
    def date_added(self) -> str:
        """Return Indicator dateAdded."""
        return self._indicator_data.get('dateAdded')  # type: ignore

    @date_added.setter
    def date_added(self, date_added: str):
        """Set Indicator dateAdded."""
        self._indicator_data['dateAdded'] = self.util.any_to_datetime(date_added).strftime(
            '%Y-%m-%dT%H:%M:%SZ'
        )

    @property
    def last_modified(self) -> str:
        """Return Indicator lastModified."""
        return self._indicator_data.get('lastModified')  # type: ignore

    @last_modified.setter
    def last_modified(self, last_modified: str):
        """Set Indicator lastModified."""
        self._indicator_data['lastModified'] = self.util.any_to_datetime(last_modified).strftime(
            '%Y-%m-%dT%H:%M:%SZ'
        )

    @property
    def first_seen(self) -> str:
        """Return Indicator firstSeen."""
        return self._indicator_data.get('firstSeen')  # type: ignore

    @first_seen.setter
    def first_seen(self, first_seen: str):
        """Set Indicator firstSeen."""
        self._indicator_data['firstSeen'] = self.util.any_to_datetime(first_seen).strftime(
            '%Y-%m-%dT%H:%M:%SZ'
        )

    @property
    def last_seen(self) -> str:
        """Return Indicator lastSeen."""
        return self._indicator_data.get('lastSeen')  # type: ignore

    @last_seen.setter
    def last_seen(self, last_seen: str):
        """Set Indicator lastSeen."""
        self._indicator_data['lastSeen'] = self.util.any_to_datetime(last_seen).strftime(
            '%Y-%m-%dT%H:%M:%SZ'
        )

    @property
    def external_date_created(self) -> str:
        """Return Indicator externalDateCreated."""
        return self._indicator_data.get('externalDateCreated')  # type: ignore

    @external_date_created.setter
    def external_date_created(self, external_date_created: str):
        """Set Indicator externalDateCreated."""
        external_date_created = self.util.any_to_datetime(external_date_created).strftime(
            '%Y-%m-%dT%H:%M:%SZ'
        )
        self._indicator_data['externalDateCreated'] = external_date_created

    @property
    def external_date_expires(self) -> str:
        """Return Indicator externalDateExpires."""
        return self._indicator_data.get('externalDateExpires')  # type: ignore

    @external_date_expires.setter
    def external_date_expires(self, external_date_expires: str):
        """Set Indicator externalDateExpires."""
        external_date_expires = self.util.any_to_datetime(external_date_expires).strftime(
            '%Y-%m-%dT%H:%M:%SZ'
        )
        self._indicator_data['externalDateExpires'] = external_date_expires

    @property
    def external_last_modified(self) -> str:
        """Return Indicator externalLastModified."""
        return self._indicator_data.get('externalLastModified')  # type: ignore

    @external_last_modified.setter
    def external_last_modified(self, external_date_last_modified: str):
        """Set Indicator externalLastModified."""
        external_date_last_modified = self.util.any_to_datetime(
            external_date_last_modified
        ).strftime('%Y-%m-%dT%H:%M:%SZ')
        self._indicator_data['externalLastModified'] = external_date_last_modified

    def occurrence(
        self,
        file_name: str | None = None,
        path: str | None = None,
        date: str | None = None,
    ) -> 'FileOccurrence | None':
        """Add a file Occurrence.

        Args:
            file_name (str, optional): The file name for this occurrence.
            path (str, optional): The file path for this occurrence.
            date (str, optional): The datetime expression for this occurrence.

        Returns:
            FileOccurrence: An instance of Occurrence.
        """
        if self._indicator_data.get('type') != 'File':
            # Indicator object has no logger to output warning
            return None

        occurrence_obj = FileOccurrence(file_name, path, date)
        self._occurrences.append(occurrence_obj)
        return occurrence_obj

    @property
    def private_flag(self) -> bool:
        """Return Indicator private flag."""
        return self._indicator_data.get('privateFlag')  # type: ignore

    @private_flag.setter
    def private_flag(self, private_flag: bool):
        """Set Indicator private flag."""
        self._indicator_data['privateFlag'] = self.util.to_bool(private_flag)

    @property
    def rating(self) -> float:
        """Return Indicator rating."""
        return self._indicator_data.get('rating')  # type: ignore

    @rating.setter
    def rating(self, rating: float):
        """Set Indicator rating."""
        self._indicator_data['rating'] = float(rating)

    @property
    def summary(self) -> str:
        """Return Indicator summary."""
        return self._indicator_data.get('summary')  # type: ignore

    def security_label(
        self, name: str, description: str | None = None, color: str | None = None
    ) -> SecurityLabel:
        """Return instance of SecurityLabel.

        .. note:: The provided security label will be create if it doesn't exist. If the security
            label already exists nothing will be changed.

        Args:
            name: The value for this security label.
            description: A description for this security label.
            color: A color (hex value) for this security label.

        Returns:
            SecurityLabel: An instance of the SecurityLabel class.
        """
        label = SecurityLabel(name, description, color)
        for label_data in self._labels:
            if label_data.name == name:
                label = label_data
                break
        else:
            self._labels.append(label)
        return label

    def tag(self, name: str, formatter: Callable[[str], str] | None = None) -> Tag:
        """Return instance of Tag.

        Args:
            name: The value for this tag.
            formatter: A method that take a tag value and returns a formatted tag.

        Returns:
            Tag: An instance of the Tag class.
        """
        tag = Tag(name, formatter)
        for tag_data in self._tags:
            if tag_data.name == name:
                tag = tag_data
                break
        else:
            self._tags.append(tag)
        return tag

    @property
    def type(self) -> str:
        """Return Group type."""
        return self._indicator_data.get('type')  # type: ignore

    @property
    def xid(self) -> str:
        """Return Group xid."""
        return self._indicator_data.get('xid')  # type: ignore

    def __str__(self) -> str:
        """Return string representation of object"""
        return json.dumps(self.data, indent=4)


class Address(Indicator):
    """ThreatConnect Batch Address Object"""

    __slots__ = []

    def __init__(self, ip: str, **kwargs):
        """Initialize instance properties.

        Args:
            ip: The value for this Indicator.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        super().__init__('Address', ip, **kwargs)


class ASN(Indicator):
    """ThreatConnect Batch ASN Object."""

    __slots__ = []

    def __init__(self, as_number: str, **kwargs):
        """Initialize instance properties.

        Args:
            as_number: The value for this Indicator.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        super().__init__('ASN', as_number, **kwargs)


class CIDR(Indicator):
    """ThreatConnect Batch CIDR Object"""

    __slots__ = []

    def __init__(self, block: str, **kwargs):
        """Initialize instance properties.

        Args:
            block: The value for this Indicator.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        super().__init__('CIDR', block, **kwargs)


class EmailAddress(Indicator):
    """ThreatConnect Batch EmailAddress Object"""

    __slots__ = []

    def __init__(self, address: str, **kwargs):
        """Initialize instance properties.

        Args:
            address: The value for this Indicator.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        super().__init__('EmailAddress', address, **kwargs)


class File(Indicator):
    """ThreatConnect Batch File Object"""

    __slots__ = []

    def __init__(
        self,
        md5: str | None = None,
        sha1: str | None = None,
        sha256: str | None = None,
        **kwargs,
    ):
        """Initialize instance properties.

        Args:
            md5: The md5 value for this Indicator.
            sha1: The sha1 value for this Indicator.
            sha256: The sha256 value for this Indicator.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            size (str, kwargs): The file size for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        summary = self.build_summary(md5, sha1, sha256)  # build the indicator summary
        super().__init__('File', summary, **kwargs)
        # self._file_action = []

    def action(self, relationship: str) -> 'FileAction':
        """Add a File Action."""
        action_obj = FileAction(self._indicator_data['xid'], relationship)  # type: ignore
        self._file_actions.append(action_obj)
        return action_obj

    @property
    def md5(self) -> str:
        """Return Indicator md5."""
        return self._indicator_data.get('md5')  # type: ignore

    @md5.setter
    def md5(self, md5: str):
        """Set Indicator md5."""
        self._indicator_data['md5'] = md5

    @property
    def sha1(self) -> str:
        """Return Indicator sha1."""
        return self._indicator_data.get('sha1')  # type: ignore

    @sha1.setter
    def sha1(self, sha1: str):
        """Set Indicator sha1."""
        self._indicator_data['sha1'] = sha1

    @property
    def sha256(self) -> str:
        """Return Indicator sha256."""
        return self._indicator_data.get('sha256')  # type: ignore

    @sha256.setter
    def sha256(self, sha256: str):
        """Set Indicator sha256."""
        self._indicator_data['sha256'] = sha256

    @property
    def size(self) -> int:
        """Return Indicator size."""
        return self._indicator_data.get('intValue1')  # type: ignore

    @size.setter
    def size(self, size: int):
        """Set Indicator size."""
        self._indicator_data['intValue1'] = size


class Host(Indicator):
    """ThreatConnect Batch Host Object"""

    __slots__ = []

    def __init__(self, hostname: str, **kwargs):
        """Initialize instance properties.

        Args:
            hostname: The value for this Indicator.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            dns_active (bool, kwargs): If True DNS active is enabled for this indicator.
            whois_active (bool, kwargs): If True WhoIs active is enabled for this indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        super().__init__('Host', hostname, **kwargs)

    @property
    def dns_active(self) -> bool:
        """Return Indicator dns active."""
        return self._indicator_data.get('flag1')  # type: ignore

    @dns_active.setter
    def dns_active(self, dns_active: bool):
        """Set Indicator dns active."""
        self._indicator_data['flag1'] = self.util.to_bool(dns_active)

    @property
    def whois_active(self) -> bool:
        """Return Indicator whois active."""
        return self._indicator_data.get('flag2')  # type: ignore

    @whois_active.setter
    def whois_active(self, whois_active: bool):
        """Set Indicator whois active."""
        self._indicator_data['flag2'] = self.util.to_bool(whois_active)


class Mutex(Indicator):
    """ThreatConnect Batch Mutex Object"""

    __slots__ = []

    def __init__(self, mutex: str, **kwargs):
        """Initialize instance properties.

        Args:
            mutex: The value for this Indicator.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        super().__init__('Mutex', mutex, **kwargs)


class RegistryKey(Indicator):
    """ThreatConnect Batch Registry Key Object"""

    __slots__ = []

    def __init__(self, key_name: str, value_name: str, value_type: str, **kwargs):
        """Initialize instance properties.

        Args:
            key_name: The key_name value for this Indicator.
            value_name: The value_name value for this Indicator.
            value_type: The value_type value for this Indicator.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        summary = self.build_summary(key_name, value_name, value_type)
        super().__init__('Registry Key', summary, **kwargs)


class URL(Indicator):
    """ThreatConnect Batch URL Object"""

    __slots__ = []

    def __init__(self, text: str, **kwargs):
        """Initialize instance properties.

        Args:
            text: The value for this Indicator.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        super().__init__('URL', text, **kwargs)


class UserAgent(Indicator):
    """ThreatConnect Batch User Agent Object"""

    __slots__ = []

    def __init__(self, text: str, **kwargs):
        """Initialize instance properties.

        Args:
            text: The value for this Indicator.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        super().__init__('User Agent', text, **kwargs)


class FileAction:
    """ThreatConnect Batch FileAction Object"""

    __slots__ = ['_action_data', '_children', 'xid']

    def __init__(self, parent_xid: str, relationship):
        """Initialize instance properties.

        .. warning:: This code is not complete and may require some update to the API.

        Args:
            parent_xid: The external id of the parent Indicator.
            relationship: ???
        """
        self.xid = str(uuid.uuid4())
        self._action_data = {
            'indicatorXid': self.xid,
            'relationship': relationship,
            'parentIndicatorXid': parent_xid,
        }
        self._children = []

    @property
    def data(self) -> dict:
        """Return File Occurrence data."""
        if self._children:
            for child in self._children:
                self._action_data.setdefault('children', []).append(child.data)
        return self._action_data

    def action(self, relationship):
        """Add a nested File Action."""
        action_obj = FileAction(self.xid, relationship)
        self._children.append(action_obj)

    def __str__(self) -> str:
        """Return string representation of object."""
        return json.dumps(self.data, indent=4)


class FileOccurrence:
    """ThreatConnect Batch FileAction Object."""

    __slots__ = ['_occurrence_data', 'util']

    def __init__(
        self,
        file_name: str | None = None,
        path: str | None = None,
        date: str | None = None,
    ):
        """Initialize instance properties

        Args:
            file_name (str, optional): The file name for this occurrence.
            path (str, optional): The file path for this occurrence.
            date (str, optional): The datetime expression for this occurrence.
        """
        self._occurrence_data = {}

        # properties
        self.util = Util()

        if file_name is not None:
            self._occurrence_data['fileName'] = file_name
        if path is not None:
            self._occurrence_data['path'] = path
        if date is not None:
            self._occurrence_data['date'] = self.util.any_to_datetime(date).strftime(
                '%Y-%m-%dT%H:%M:%SZ'
            )

    @property
    def data(self) -> dict:
        """Return File Occurrence data."""
        return self._occurrence_data

    @property
    def date(self) -> str:
        """Return File Occurrence date."""
        return self._occurrence_data.get('date')  # type: ignore

    @date.setter
    def date(self, date: str):
        """Set File Occurrence date."""
        self._occurrence_data['date'] = self.util.any_to_datetime(date).strftime(
            '%Y-%m-%dT%H:%M:%SZ'
        )

    @property
    def file_name(self) -> str:
        """Return File Occurrence file name."""
        return self._occurrence_data.get('fileName')  # type: ignore

    @file_name.setter
    def file_name(self, file_name: str):
        """Set File Occurrence file name."""
        self._occurrence_data['fileName'] = file_name

    @property
    def path(self) -> str:
        """Return File Occurrence path."""
        return self._occurrence_data.get('path')  # type: ignore

    @path.setter
    def path(self, path: str):
        """Set File Occurrence path."""
        self._occurrence_data['path'] = path

    def __str__(self) -> str:
        """Return string representation of object."""
        return json.dumps(self.data, indent=4)



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class WorkflowTemplateModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='WorkflowTemplate Model',
    validate_assignment=True,
):
    """Workflow_Template Model"""

    _associated_type = PrivateAttr(default=False)
    _cm_type = PrivateAttr(default=True)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    active: bool = Field(
        None,
        allow_mutation=False,
        description='The **active** for the Workflow_Template.',
        read_only=True,
        title='active',
    )
    assignee: 'AssigneeModel' = Field(
        None,
        allow_mutation=False,
        description='The **assignee** for the Workflow_Template.',
        read_only=True,
        title='assignee',
    )
    cases: 'CasesModel' = Field(
        None,
        allow_mutation=False,
        description='The **cases** for the Workflow_Template.',
        read_only=True,
        title='cases',
    )
    config_artifact: str | None = Field(
        None,
        allow_mutation=False,
        description='The **config artifact** for the Workflow_Template.',
        read_only=True,
        title='configArtifact',
    )
    config_attribute: dict | list[dict] | None = Field(
        None,
        description='The **config attribute** for the Workflow_Template.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='configAttribute',
    )
    config_playbook: str | None = Field(
        None,
        allow_mutation=False,
        description='The **config playbook** for the Workflow_Template.',
        read_only=True,
        title='configPlaybook',
    )
    config_task: dict | list[dict] | None = Field(
        None,
        allow_mutation=False,
        description='The **config task** for the Workflow_Template.',
        read_only=True,
        title='configTask',
    )
    description: str | None = Field(
        None,
        description='The **description** for the Workflow_Template.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='description',
    )
    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    name: str | None = Field(
        None,
        description='The **name** for the Workflow_Template.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='name',
    )
    owner: str | None = Field(
        None,
        allow_mutation=False,
        description='The name of the Owner of the Case.',
        read_only=True,
        title='owner',
    )
    owner_id: int | None = Field(
        None,
        allow_mutation=False,
        description='The name of the Owner of the Case.',
        read_only=True,
        title='ownerId',
    )
    version: int | None = Field(
        None,
        description='The **version** for the Workflow_Template.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='version',
    )

    @validator('assignee', always=True, pre=True)
    @classmethod
    def _validate_assignee(cls, v):
        if not v:
            return AssigneeModel()  # type: ignore
        return v

    @validator('cases', always=True, pre=True)
    @classmethod
    def _validate_cases(cls, v):
        if not v:
            return CasesModel()  # type: ignore
        return v


class WorkflowTemplateDataModel(
    BaseModel,
    title='WorkflowTemplate Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Workflow_Templates Data Model"""

    data: list[WorkflowTemplateModel] | None = Field(
        [],
        description='The data for the WorkflowTemplates.',
        methods=['POST', 'PUT'],
        title='data',
    )


class WorkflowTemplatesModel(
    BaseModel,
    title='WorkflowTemplates Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Workflow_Templates Model"""

    _mode_support = PrivateAttr(default=False)

    data: list[WorkflowTemplateModel] | None = Field(
        [],
        description='The data for the WorkflowTemplates.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.cases.case_model import CasesModel
from tcex.api.tc.v3.security.assignee_model import AssigneeModel

# add forward references
WorkflowTemplateDataModel.update_forward_refs()
WorkflowTemplateModel.update_forward_refs()
WorkflowTemplatesModel.update_forward_refs()



"""TcEx Framework Module"""

# standard library
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..app.playbook.playbook import Playbook  # type: ignore


class PlaybookOutput(dict):
    """Playbook Output

    Args:
        playbook: An instance of Playbook.
    """

    def __init__(self, playbook: 'Playbook'):
        """Initialize the class properties."""
        super().__init__()
        self.playbook = playbook

    def process(self):
        """Create all stored output data to storage."""
        for key, value in self.items():
            self.playbook.create.variable(key, value)



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class SystemRoleModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='SystemRole Model',
    validate_assignment=True,
):
    """System_Role Model"""

    _associated_type = PrivateAttr(default=False)
    _cm_type = PrivateAttr(default=False)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    active: bool = Field(
        None,
        allow_mutation=False,
        description='The **active** for the System_Role.',
        read_only=True,
        title='active',
    )
    assignable: bool = Field(
        None,
        allow_mutation=False,
        description='The **assignable** for the System_Role.',
        read_only=True,
        title='assignable',
    )
    displayed: bool = Field(
        None,
        allow_mutation=False,
        description='The **displayed** for the System_Role.',
        read_only=True,
        title='displayed',
    )
    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    name: str | None = Field(
        None,
        allow_mutation=False,
        description='The **name** for the System_Role.',
        read_only=True,
        title='name',
    )


class SystemRoleDataModel(
    BaseModel,
    title='SystemRole Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """System_Roles Data Model"""

    data: list[SystemRoleModel] | None = Field(
        [],
        description='The data for the SystemRoles.',
        methods=['POST', 'PUT'],
        title='data',
    )


class SystemRolesModel(
    BaseModel,
    title='SystemRoles Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """System_Roles Model"""

    _mode_support = PrivateAttr(default=False)

    data: list[SystemRoleModel] | None = Field(
        [],
        description='The data for the SystemRoles.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# add forward references
SystemRoleDataModel.update_forward_refs()
SystemRoleModel.update_forward_refs()
SystemRolesModel.update_forward_refs()



"""TcEx Framework Module"""

# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql import Tql
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.tql.tql_type import TqlType


class TagFilter(FilterABC):
    """Filter Object for Tags"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.TAGS.value

    def associated_case(self, operator: Enum, associated_case: int | list):
        """Filter associatedCase based on **associatedCase** keyword.

        Args:
            operator: The operator enum for the filter.
            associated_case: No description provided.
        """
        if isinstance(associated_case, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('associatedCase', operator, associated_case, TqlType.INTEGER)

    def associated_group(self, operator: Enum, associated_group: int | list):
        """Filter associatedGroup based on **associatedGroup** keyword.

        Args:
            operator: The operator enum for the filter.
            associated_group: No description provided.
        """
        if isinstance(associated_group, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('associatedGroup', operator, associated_group, TqlType.INTEGER)

    def associated_indicator(self, operator: Enum, associated_indicator: int | list):
        """Filter associatedIndicator based on **associatedIndicator** keyword.

        Args:
            operator: The operator enum for the filter.
            associated_indicator: No description provided.
        """
        if isinstance(associated_indicator, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('associatedIndicator', operator, associated_indicator, TqlType.INTEGER)

    def associated_victim(self, operator: Enum, associated_victim: int | list):
        """Filter associatedVictim based on **associatedVictim** keyword.

        Args:
            operator: The operator enum for the filter.
            associated_victim: No description provided.
        """
        if isinstance(associated_victim, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('associatedVictim', operator, associated_victim, TqlType.INTEGER)

    def case_id(self, operator: Enum, case_id: int | list):
        """Filter Case ID based on **caseId** keyword.

        Args:
            operator: The operator enum for the filter.
            case_id: The ID of the case the tag is applied to.
        """
        if isinstance(case_id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('caseId', operator, case_id, TqlType.INTEGER)

    def description(self, operator: Enum, description: list | str):
        """Filter Description based on **description** keyword.

        Args:
            operator: The operator enum for the filter.
            description: The description of the tag.
        """
        if isinstance(description, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('description', operator, description, TqlType.STRING)

    @property
    def has_case(self):
        """Return **CaseFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.cases.case_filter import CaseFilter

        cases = CaseFilter(Tql())
        self._tql.add_filter('hasCase', TqlOperator.EQ, cases, TqlType.SUB_QUERY)
        return cases

    @property
    def has_group(self):
        """Return **GroupFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.groups.group_filter import GroupFilter

        groups = GroupFilter(Tql())
        self._tql.add_filter('hasGroup', TqlOperator.EQ, groups, TqlType.SUB_QUERY)
        return groups

    @property
    def has_indicator(self):
        """Return **IndicatorFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.indicators.indicator_filter import IndicatorFilter

        indicators = IndicatorFilter(Tql())
        self._tql.add_filter('hasIndicator', TqlOperator.EQ, indicators, TqlType.SUB_QUERY)
        return indicators

    @property
    def has_victim(self):
        """Return **VictimFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.victims.victim_filter import VictimFilter

        victims = VictimFilter(Tql())
        self._tql.add_filter('hasVictim', TqlOperator.EQ, victims, TqlType.SUB_QUERY)
        return victims

    def id(self, operator: Enum, id: int | list):  # noqa: A002
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the tag.
        """
        if isinstance(id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def last_used(self, operator: Enum, last_used: Arrow | datetime | int | str):
        """Filter LastUsed based on **lastUsed** keyword.

        Args:
            operator: The operator enum for the filter.
            last_used: The date this tag was last used.
        """
        last_used = self.util.any_to_datetime(last_used).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('lastUsed', operator, last_used, TqlType.STRING)

    def name(self, operator: Enum, name: list | str):
        """Filter Name based on **name** keyword.

        Args:
            operator: The operator enum for the filter.
            name: The name of the tag (case sensitive).
        """
        if isinstance(name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('name', operator, name, TqlType.STRING)

    def owner(self, operator: Enum, owner: int | list):
        """Filter Owner ID based on **owner** keyword.

        Args:
            operator: The operator enum for the filter.
            owner: The owner ID of the tag.
        """
        if isinstance(owner, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('owner', operator, owner, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The owner name of the tag.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def security_coverage(self, operator: Enum, security_coverage: list | str):
        """Filter Security Coverage based on **securityCoverage** keyword.

        Args:
            operator: The operator enum for the filter.
            security_coverage: The security coverage level of an ATT&CK-based tag as it relates to
                the user's organization.
        """
        if isinstance(security_coverage, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('securityCoverage', operator, security_coverage, TqlType.STRING)

    def summary(self, operator: Enum, summary: list | str):
        """Filter Summary based on **summary** keyword.

        Args:
            operator: The operator enum for the filter.
            summary: The name of the tag (case insensitive).
        """
        if isinstance(summary, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('summary', operator, summary, TqlType.STRING)

    def technique_id(self, operator: Enum, technique_id: list | str):
        """Filter Technique ID based on **techniqueId** keyword.

        Args:
            operator: The operator enum for the filter.
            technique_id: The standard ID for specific MITRE ATT&CK techniques and subtechniques.
        """
        if isinstance(technique_id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('techniqueId', operator, technique_id, TqlType.STRING)
